from stoobly_orator.orm import belongs_to, has_many

from .base import Base
from .trace_alias import TraceAlias

class TraceRequest(Base):
  __fillable__ = ['trace_id']

  @belongs_to
  def trace(self):
    from .trace import Trace
    return Trace

  @has_many
  def trace_aliases(self):
    return TraceAlias

def handle_deleting(trace):
  trace_aliaces = trace.trace_aliases

  for trace_alias in trace_aliaces:
    trace_alias.delete()

TraceRequest.deleting(handle_deleting)


import json
import pdb

from typing import List

from ..types.print_options import PrintOptions

def json_print(records: List[dict], **kwargs: PrintOptions):
  filter = kwargs.get('filter') or []
  print_handler = kwargs.get('print_handler') or print
  select = kwargs.get('select') or []

  rows = []
  for record in records:
    row = {}
    for key, val in record.items():
      # If exclude list is provided, check if key is excluded
      if key in filter:
        continue
      
      # If include list is provided, check if key is included
      if len(select) > 0 and key not in select:
        continue

      row[key] = str(val) if val != None else None
    
    rows.append(row)

  print_handler(json.dumps(rows, indent=2))


import pdb
import requests

from stoobly_orator.orm import belongs_to

from stoobly_agent.app.proxy.mitmproxy.response_facade import MitmproxyResponseFacade
from stoobly_agent.app.proxy.record.response_string import ResponseString
from stoobly_agent.app.models.adapters.python import PythonResponseAdapterFactory

from .base import Base
from .request import Request

class ReplayedResponse(Base):
  __fillable__ = ['latency', 'timestamp', 'raw', 'status', 'request_id']

  def with_python_response(self, response: requests.Response):
    mitmproxy_response = PythonResponseAdapterFactory(response).mitmproxy_response()
    adapted_response = MitmproxyResponseFacade(mitmproxy_response)
    response_string = ResponseString(adapted_response, None) 
    self.raw = response_string.get()

    self.status = response.status_code

  @belongs_to
  def request(self):
    return Request

  # Override
  def to_dict(self):
    return super().attributes_to_dict()


import pdb

from requests import Request

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter

from ..request import Request as ORMRequest

class ORMToRequestTransformer():

  def __init__(self, request: ORMRequest):
    self.__dirty = False

    self.__body = None
    self.__headers = None
    self.__method = None
    self.__request = request
    self.__url = None

  def transform(self) -> Request:
    adapter = RawHttpRequestAdapter(self.__request.raw)

    if self.__body:
      adapter.body = self.__body 

    if self.__headers:
      adapter.headers = self.__headers

    if self.__method:
      adapter.method = self.__method

    if self.__url:
      adapter.url = self.__url

    python_request = adapter.to_request()

    return python_request

  @property
  def dirty(self):
    return self.__dirty

  def with_body(self, body: str):
    self.__body = body
    self.__dirty = True
    return self

  def with_headers(self, headers: list):
    self.__headers = headers
    self.__dirty = True
    return self

  def with_method(self, method: str):
    _method = method.strip().upper()
    if not self.__validate_method(_method):
      return self
    self.__method = _method
    self.__dirty = True
    return self

  def with_url(self, url: str):
    self.__url = url
    self.__dirty = True
    return self

  def __validate_method(self, method: str):
    if not isinstance(method, str):
      return False

    return method.upper() in ['CONNECT', 'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PATCH', 'POST', 'PUT', 'TRACE']


import re

from stoobly_agent.app.api.bodies_controller import BodiesController
from stoobly_agent.app.api.headers_controller import HeadersController
from stoobly_agent.app.api.query_params_controller import QueryParamsController
from stoobly_agent.app.api.replayed_responses_controller import ReplayedResponsesController
from stoobly_agent.app.api.requests_controller import RequestsController
from stoobly_agent.app.api.responses_controller import ResponsesController
from stoobly_agent.app.api.response_headers_controller import ResponseHeadersController
from stoobly_agent.app.api.scenarios_controller import ScenariosController

from .configs_controller import ConfigsController
from .proxy_controller import ProxyController
from .replayed_response_headers_controller import ReplayedResponseHeadersController
from .statuses_controller import StatusesController

CONFIGS_PATH = '/configs'
PROXY_PATH = '/proxy'
REQUESTS_PATH = '/requests'
SCENARIOS_PATH = '/scenarios'
STATUSES_PATH = '/statuses'

BODIES_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/bodies")
BODY_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/bodies/[^/]+")
HEADERS_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/headers")
HEADER_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/headers/[^/]+")
RESPONSE_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/responses/[^/]+")
RESPONSES_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/responses")
QUERY_PARAMS_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/query_params")
QUERY_PARAM_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/query_params/[^/]+")
REPLAYED_RESPONSES_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/replayed_responses")
REPLAYED_RESPONSE_PATH = re.compile(f"{REPLAYED_RESPONSES_PATH.pattern}/[^/]+")
REPLAYED_RESPONSE_HEADERS_PATH = re.compile(f"{REPLAYED_RESPONSES_PATH.pattern}/[^/]+/headers")
RESPONSE_HEADERS_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/response_headers")
RESPONSE_HEADER_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+/response_headers/[^/]+")
REQUEST_PATH = re.compile(f"{REQUESTS_PATH}/[^/]+")
SCENARIO_PATH = re.compile(f"{SCENARIOS_PATH}/[^/]+")
STATUS_PATH = re.compile(f"{STATUSES_PATH}/[^/]+")

ROUTES = {
  'DELETE': [
    [HEADERS_PATH, HeadersController.instance().destroy],
    [QUERY_PARAMS_PATH, QueryParamsController.instance().destroy],
    [RESPONSE_HEADERS_PATH, ResponseHeadersController.instance().destroy],
    [REQUEST_PATH, RequestsController.instance().destroy],
    [SCENARIO_PATH, ScenariosController.instance().destroy],
  ],
  'GET': [
      [CONFIGS_PATH, ConfigsController.instance().show],
      ['/'.join([CONFIGS_PATH, 'summary']), ConfigsController.instance().summary],
      ['/'.join([CONFIGS_PATH, 'policies']), ConfigsController.instance().policies],
      ['/'.join([PROXY_PATH, 'get']), ProxyController.instance().do_GET],
      [re.compile('/'.join([BODIES_PATH.pattern, 'mock'])), BodiesController.instance().mock],
      [HEADERS_PATH, HeadersController.instance().index],
      [QUERY_PARAMS_PATH, QueryParamsController.instance().index],
      [REPLAYED_RESPONSE_HEADERS_PATH, ReplayedResponseHeadersController().index],
      [re.compile('/'.join([REPLAYED_RESPONSE_PATH.pattern, 'mock'])), ReplayedResponsesController().mock],
      [REPLAYED_RESPONSES_PATH, ReplayedResponsesController().index],
      [re.compile('/'.join([RESPONSES_PATH.pattern, 'mock'])), ResponsesController.instance().mock],
      [RESPONSES_PATH, ResponsesController.instance().show],
      [RESPONSE_HEADERS_PATH, ResponseHeadersController.instance().index],
      [REQUESTS_PATH, RequestsController.instance().index],
      [re.compile('/'.join([REQUEST_PATH.pattern, 'replay'])), RequestsController.instance().replay],
      [re.compile('/'.join([REQUEST_PATH.pattern, 'download$'])), RequestsController.instance().download],
      [REQUEST_PATH, RequestsController.instance().get],
      [SCENARIOS_PATH, ScenariosController.instance().index],
      [re.compile('/'.join([SCENARIO_PATH.pattern, 'download$'])), ScenariosController.instance().download],
      [STATUSES_PATH, StatusesController.instance().index],
      [SCENARIO_PATH, ScenariosController.instance().get],
  ],
  'POST': [
      ['/'.join([PROXY_PATH, 'post']), ProxyController.instance().do_POST],
      [HEADERS_PATH, HeadersController.instance().create],
      [QUERY_PARAMS_PATH, QueryParamsController.instance().create],
      [RESPONSE_HEADERS_PATH, ResponseHeadersController.instance().create],
      [REQUESTS_PATH, RequestsController.instance().create],
      [re.compile('/'.join([REQUEST_PATH.pattern, 'push$'])), RequestsController.instance().push],
      [SCENARIOS_PATH, ScenariosController.instance().create],
  ],
  'PUT': [
      [CONFIGS_PATH, ConfigsController.instance().update],
      ['/'.join([PROXY_PATH, 'put']), ProxyController.instance().do_PUT],
      [BODY_PATH, BodiesController.instance().update],
      [HEADER_PATH, HeadersController.instance().update],
      [QUERY_PARAM_PATH, QueryParamsController.instance().update],
      [RESPONSE_PATH, ResponsesController.instance().update],
      [RESPONSE_HEADER_PATH, ResponseHeadersController.instance().update],
      [re.compile('/'.join([REPLAYED_RESPONSE_PATH.pattern, 'activate'])), ReplayedResponsesController().activate],
      ['/'.join([REQUESTS_PATH, 'send']), RequestsController.instance().send],
      [REQUEST_PATH, RequestsController.instance().update],
      [STATUS_PATH, StatusesController.instance().update],
      [SCENARIO_PATH, ScenariosController.instance().update],
  ]
}


from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.lib.api.scenarios_resource import ScenariosResource
from stoobly_agent.app.settings import RemoteSettings

from .local_db.scenario_adapter import LocalDBScenarioAdapter
from .stoobly.scenario_adapter import StooblyScenarioAdapter

class ScenarioResourceFactory():

  def __init__(self, settings: RemoteSettings):
    self.__remote_settings = settings

  def local_db(self) -> LocalDBScenarioAdapter:
    return LocalDBScenarioAdapter(Scenario)  

  def stoobly(self) -> StooblyScenarioAdapter:
    api = ScenariosResource(self.__remote_settings.api_url, self.__remote_settings.api_key)
    return StooblyScenarioAdapter(api)


from urllib.request import Request
from stoobly_agent.lib.api.interfaces.request_show_response import RequestShowResponse
from stoobly_agent.lib.api.interfaces.response_show_response import ResponseShowResponse

from ..context_response import TestContextResponse

class StooblyResponseAdapter():

  def __init__(self, request: RequestShowResponse, response: ResponseShowResponse) -> None:
      self._request = request
      self._response = response

  def adapt(self):
    response = TestContextResponse()

    response.with_content(self._response['text'])
    response.with_status_code(self._request['status'])

    return response


import re
import time
import pdb

from stoobly_agent.app.proxy.mitmproxy.response_facade import MitmproxyResponseFacade

from .response_string_control import ResponseStringControl

CLRF = b"\r\n"

class ResponseString:
    ENCODING = 'utf-8'
    RESPONSE_TYPE = 2

    __current_time = None

    def __init__(self, response: MitmproxyResponseFacade, request_id):
        self.__current_time = self.__get_current_time()

        self.response = response
        self.lines = []
        self.latency = 0
        self.request_id = request_id

        if response:
            self.__response_line()
            self.__headers()
            self.__body()

    def get(self, **kwargs):
        if kwargs.get('control'):
            return CLRF.join([self.control] + self.lines)
        else:
            return CLRF.join(self.lines)

    def set(self, s: bytes):
        if not isinstance(s, bytes):
            s = s.encode(self.ENCODING)

        self.lines = s.split(CLRF)

    ###
    #
    # 1 - response type
    # 2 - request id
    # 3 - timestamp in nano seconds
    # 4 - response latency in nano seconds
    #
    @property
    def control(self):
        control = ResponseStringControl()
        control.id = self.request_id
        control.timestamp = self.__current_time
        control.latency = self.latency

        return control.serialize().encode(self.ENCODING)

    @control.setter
    def control(self, c):
        control = ResponseStringControl(c)
        self.request_id = control.id
        self.__current_time = control.timestamp
        self.latency = control.latency

    def with_latency(self, latency):
        self.latency = latency
        return self

    def with_request_id(self, request_id):
        self.request_id = request_id
        return self

    def __response_line(self):
        line = "HTTP/1.1 {}".format(self.response.code)
        self.lines.append(line.encode(self.ENCODING))

    def __headers(self):
        headers = self.response.headers

        for name, val in headers.items():
            line = "{}: {}".format(self.__to_header_case(name), val)
            self.lines.append(line.encode(self.ENCODING))

    def __body(self):
        body = self.response.body

        if not body:
            self.lines.append(CLRF)
        elif isinstance(body, bytes):
            self.lines.append(CLRF + body)
        else:
            raise Exception('Unsupported body type')

    def __to_header_case(self, header):
        toks = re.split('_|-', header)
        if len(toks) == 1:
            return header

        for index, tok in enumerate(toks):
            toks[index] = tok.lower().capitalize()
        return "-".join(toks)

    def __get_current_time(self):
        now = time.time()
        current_time = round(now * (pow(10, 6)))
        return current_time



import uuid

from mitmproxy.http import Response as MitmproxyResponse

class LocalDBAdapter():

  def content_encoding_header(self, new_headers: dict, old_headers: dict):
    content_encoding = new_headers.get('content-encoding')
    if not content_encoding:
      # If content-encoding has changed, we will need to decode 'text' and re-encode it
      content_encoding = old_headers.get('content-encoding') 
    return content_encoding

  def encode_body(self, body: str, content_encoding: str):
    # If a content encoding is set in the response headers, encode the body
    # TODO: we may want to create a separate endpoint for this and have the client drive encoding before updating
    # otherwise, there's an assumption that the 'text' will always be provided as decoded

    if content_encoding:
      # Use MitmproxyResponse to encode the body
      res = MitmproxyResponse.make(
          200,
          body,
          {'content-encoding': content_encoding}
      )
      return res.raw_content
    else: 
      return body.encode()

  def success(self, d):
    return d, 200

  def bad_request(self, d = ''):
    return d, 400

  def not_found(self, d = 'Not Found'):
    return d, 404

  def internal_error(self, d = 'Internal Error'):
    return d, 500

  def validate_uuid(self, id: str):
    try:
      uuid.UUID(id)
      return True
    except Exception:
      return False


import os
import pdb
import yaml

from typing import List

from stoobly_agent.config.data_dir import DATA_DIR_NAME, TMP_DIR_NAME
from stoobly_agent.app.cli.scaffold.constants import APP_DIR
from stoobly_agent.app.cli.scaffold.service_config import ServiceConfig
from stoobly_agent.app.cli.scaffold.docker.constants import APP_INGRESS_NETWORK_NAME, APP_EGRESS_NETWORK_NAME, DOCKER_COMPOSE_BASE, DOCKER_COMPOSE_BASE_TEMPLATE
from stoobly_agent.app.cli.scaffold.templates.constants import CORE_GATEWAY_SERVICE_NAME

def configure_gateway(service_paths: List[str], no_publish = False):
  if len(service_paths) == 0:
    return

  hostnames, ports = __find_hosts(service_paths)

  service_dir_path = os.path.dirname(service_paths[0])
  gateway_service_path = os.path.join(service_dir_path, CORE_GATEWAY_SERVICE_NAME)
  docker_compose_src_path = os.path.join(gateway_service_path, DOCKER_COMPOSE_BASE_TEMPLATE)
  docker_compose_dest_path = os.path.join(gateway_service_path, DOCKER_COMPOSE_BASE)

  compose = {}
  with open(docker_compose_src_path, 'r') as fp:
    compose = yaml.safe_load(fp)
    services = compose.get('services')

    if not services:
      return
    
    gateway_base = services.get('gateway_base')

    if gateway_base:
      if not no_publish:
        gateway_base['ports'] = ports

      app_dir_path = os.path.dirname(os.path.dirname(service_dir_path))
      __with_traefik_config(service_paths, gateway_base, app_dir_path)
      __with_networks(gateway_base, hostnames) 

  with open(docker_compose_dest_path, 'w') as fp:
    yaml.dump(compose, fp) 

def __with_networks(config: dict, hostnames: List[str]):
  networks = {} 
  config['networks'] = networks
  networks[APP_EGRESS_NETWORK_NAME] = {}
  networks[APP_INGRESS_NETWORK_NAME] = {
    'aliases': hostnames
  }

def __with_traefik_config(service_paths: str, compose: dict, app_dir_path: str):
  if not compose['volumes']:
    compose['volumes'] = []

  entry_points = {}
  certificates = []
  traefik_config = {
    'accessLog': {
      'format': 'common',
    },
    'entryPoints': entry_points,
    'log': {
      'format': 'common',
      'level': 'INFO',
    },
    'providers': {
      'docker': {
        'exposedByDefault': False
      }
    },
    'tls': {
      'certificates': certificates
    }
  }

  for path in service_paths:
    config = ServiceConfig(path)

    if not config.hostname:
      continue

    entry_points[config.port] = {
      'address': f":{config.port}"
    }

    if config.scheme == 'https':
      certificates.append({
        'certFile': f"/certs/{config.hostname}.crt",
        'keyFile': f"/certs/{config.hostname}.key"
      })

  # Create traefik.yml in .stoobly/tmp
  traefik_template_relative_path = os.path.join(DATA_DIR_NAME, TMP_DIR_NAME, 'traefik.yml')
  traefik_template_dest_path = os.path.join(app_dir_path, traefik_template_relative_path)

  if not os.path.exists(os.path.dirname(traefik_template_dest_path)):
    os.makedirs(os.path.dirname(traefik_template_dest_path), exist_ok=True)

  with open(traefik_template_dest_path, 'w') as fp:
    fp.write(yaml.dump(traefik_config))

  compose['volumes'].append(
    f"{os.path.join(APP_DIR, traefik_template_relative_path)}:/etc/traefik/traefik.yml:ro"
  )

def __find_hosts(service_paths):
  hostnames = []
  ports = []
  for path in service_paths:
    config = ServiceConfig(path)

    if not config.hostname:
      continue

    try:
      port = int(config.port)
    except Exception:
      continue
    
    port_mapping = f"{port}:{port}"
    if port > 0 and port <= 65535 and port_mapping not in ports:
      ports.append(port_mapping)

    if config.hostname not in hostnames:
      hostnames.append(config.hostname) 

  return hostnames, ports


import os
import pdb
import shutil

from .app import App
from .constants import WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE
from .docker.service.builder import ServiceBuilder
from .docker.workflow.decorators_factory import get_workflow_decorators
from .service_command import ServiceCommand
from .workflow_create_command import WorkflowCreateCommand

class ServiceCreateCommand(ServiceCommand):

  def __init__(self, app: App, **kwargs):
    super().__init__(app, **kwargs)

    self.__env_vars = kwargs.get('env') or []
    self.__workflows = kwargs.get('workflow') or []

  @property
  def env_vars(self):
    return self.__env_vars

  @property
  def workflows(self):
    return self.__workflows

  def build(self):
    service_builder = ServiceBuilder(self.service_config)
    service_builder.with_env(list(self.env_vars))
    service_decorators = []

    for service_decorator in service_decorators:
      service_decorator(service_builder).decorate()

    service_builder.write()

    workflow_kwargs = {
      'app_dir_path': self.app_dir_path,
      'namespace': self.namespace, 
      'service_name': self.service_name,
    }

    if WORKFLOW_MOCK_TYPE in self.workflows:
      self.__build_with_mock_workflow(service_builder, **workflow_kwargs)

    if WORKFLOW_RECORD_TYPE in self.workflows:
      self.__build_with_record_workflow(service_builder, **workflow_kwargs)

    if WORKFLOW_TEST_TYPE in self.workflows:
      self.__build_with_test_workflow(service_builder, **workflow_kwargs)

    self.service_config.write()

  def reset(self):
    dest = self.service_path

    if os.path.exists(dest):
      shutil.rmtree(dest)

  def __build_with_mock_workflow(self, service_builder: ServiceBuilder, **kwargs):
    mock_workflow = WorkflowCreateCommand(self.app, **{ **kwargs, **{ 'workflow_name': WORKFLOW_MOCK_TYPE }})

    workflow_decorators = get_workflow_decorators(WORKFLOW_MOCK_TYPE, self.service_config)
    mock_workflow.build(service_builder=service_builder, workflow_decorators=workflow_decorators)

  def __build_with_record_workflow(self, service_builder: ServiceBuilder, **kwargs):
    record_workflow = WorkflowCreateCommand(self.app, **{ **kwargs, **{ 'workflow_name': WORKFLOW_RECORD_TYPE }})

    workflow_decorators = get_workflow_decorators(WORKFLOW_RECORD_TYPE, self.service_config)
    record_workflow.build(service_builder=service_builder, workflow_decorators=workflow_decorators)

  def __build_with_test_workflow(self, service_builder: ServiceBuilder, **kwargs):
    mock_workflow = WorkflowCreateCommand(self.app, **{ **kwargs, **{ 'workflow_name': WORKFLOW_TEST_TYPE }})

    workflow_decorators = get_workflow_decorators(WORKFLOW_TEST_TYPE, self.service_config)
    mock_workflow.build(service_builder=service_builder, workflow_decorators=workflow_decorators)


import pdb
import requests

from .python_adapter import PythonResponseAdapter

class MitmproxyResponseAdapterFactory():

  def __init__(self, response: requests.Response):
    self.__response = response

  def python_response(self):
    return PythonResponseAdapter(self.__response).adapt()


import pdb
import requests

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.models.body_model import BodyModel
from stoobly_agent.app.settings import Settings

class BodiesController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # PUT /requests/:requestId/bodies/:bodyId
    def update(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        text = context.params.get('text')
        if text == None:
            return context.bad_request('Missing text')

        body_model = self.__body_model(context)
        request, status = body_model.update(context.params.get('requestId'), text)

        if context.filter_response(request, status):
            return

        context.render(
            json = {
                'text': text,
            },
            status = 200 
        )

    # GET /requests/:requestId/bodies/mock
    def mock(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        body_model = self.__body_model(context)
        request, status = body_model.mock(context.params.get('requestId'))

        if context.filter_response(request, status):
            return

        # Extract specific headers
        headers = {}

        accepted_headers = ['content-encoding', 'content-type']
        for header, val in request.headers.items():
            decoded_header = header.lower()

            if decoded_header not in accepted_headers:
                continue 

            headers[decoded_header] = val

        context.render(
            data = request.data,
            headers = headers,
            status = 200
        )

    def __body_model(self, context: SimpleHTTPRequestHandler) -> BodyModel:
        access_token = context.headers.get('access-token')
        body_model = BodyModel(Settings.instance(), access_token=access_token)
        return body_model


import os

from .builder import Builder
from .constants import DOCKER_COMPOSE_BASE, DOCKERFILE_CONTEXT
from ..app_config import AppConfig

class AppBuilder(Builder):

  def __init__(self, config: AppConfig):
    super().__init__(config.dir, DOCKER_COMPOSE_BASE)

  @property
  def context_base(self):
    return 'context_base'

  @property
  def context_docker_file_path(self):
    return os.path.join(self.dir_path, DOCKERFILE_CONTEXT)

  @property
  def stoobly_base(self):
    return 'stoobly_base'


import pdb
import requests


from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.models.response_model import ResponseModel
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import custom_headers

class ResponsesController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # GET /requests/:requestId/responses
    def show(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        response_model = self.__response_model(context) 
        response, status = response_model.show(context.params.get('requestId'))
        
        if context.filter_response(response, status):
            return

        return context.render(
            json = response,
            status = 200
        )

    # PUT /requests/:requestId/responses/:responseId
    def update(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1,
            'responseId': 3,
        })
 
        response_model = self.__response_model(context) 
        response, status = response_model.update(context.params.get('responseId'), **context.params)

        if context.filter_response(response, status):
            return

        return context.render(
            json = response,
            status = 200
        )

    # GET /requests/:requestId/responses/mock
    def mock(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        response_model = self.__response_model(context) 
        response, status = response_model.mock(context.params.get('requestId'))

        if context.filter_response(response, status):
            return

        return self.__render_response(context, response)

    def __response_model(self, context: SimpleHTTPRequestHandler):
        access_token = context.headers.get('access-token')
        response_model = ResponseModel(Settings.instance(), access_token=access_token)
        return response_model

    def __render_response(self, context: SimpleHTTPRequestHandler, response: requests.Response):
        if response == None:
            return context.render(
                plain = '',
                status = 404
            )

        # Extract specific headers
        headers = {}

        accepted_headers = ['content-type', custom_headers.RESPONSE_ID.lower()]
        for header, val in response.headers.items():
            decoded_header = header.lower()

            if decoded_header not in accepted_headers:
                continue 

            headers[decoded_header] = val

        expose_headers = 'Access-Control-Expose-Headers'
        if expose_headers in headers:
            headers[expose_headers] = ','.join([headers[expose_headers], custom_headers.RESPONSE_ID])
        else:
            headers[expose_headers] = custom_headers.RESPONSE_ID

        context.render(
            data = response.content,
            headers = headers,
            status = 200
        )


import pdb
import os
import shutil

from stoobly_agent.config.constants.env_vars import ENV

CERTS_DIR_NAME = 'certs'
DATA_DIR_NAME = '.stoobly'
DB_FILE_NAME = 'stoobly_agent.sqlite3'
DB_VERSION_NAME = 'VERSION'
MITMPROXY_OPTIONS_FILE_NAME = 'options.json'
TMP_DIR_NAME = 'tmp'

class DataDir:

    _instances = None

    def __init__(self, path: str = None):
        if DataDir._instances.get(path):
            raise RuntimeError('Call instance() instead')
        else:
            self.__path = path

            if path:
                self.__data_dir_path = os.path.join(path, DATA_DIR_NAME)
            else:
                cwd = os.getcwd()
                self.__data_dir_path = self.find_data_dir(cwd)

            if not os.path.exists(self.__data_dir_path):
                self.create(os.path.dirname(self.__data_dir_path))

    def __repr__(self) -> str:
      return self.path

    @classmethod
    def instance(cls, path: str = None):
        if not cls._instances:
            cls._instances = {}

        if not cls._instances.get(path):
            cls._instances[path] = cls(path)

        return cls._instances[path]

    @classmethod
    def handle_chdir(cls):
        if cls._instances and None in cls._instances:
            del cls._instances[None]

        return cls.instance()

    @property
    def context_dir_path(self):
        return os.path.abspath(os.path.join(self.path, '..'))

    @property
    def path(self):
        if not self.__path and os.environ.get(ENV) == 'test':
            test_path = os.path.join(self.__data_dir_path, TMP_DIR_NAME, DATA_DIR_NAME)

            if not os.path.exists(test_path):
                os.makedirs(test_path, exist_ok=True)

            return test_path

        return self.__data_dir_path

    @property
    def tmp_dir_path(self):
        tmp_dir_path = os.path.join(self.path, TMP_DIR_NAME)

        if not os.path.exists(tmp_dir_path):
            os.mkdir(tmp_dir_path)

        return tmp_dir_path

    @property
    def certs_dir_path(self):
        certs_dir_path = os.path.join(self.path, CERTS_DIR_NAME)

        if not os.path.exists(certs_dir_path):
            os.mkdir(certs_dir_path)

        return certs_dir_path

    @property
    def db_dir_path(self):
        db_dir_path = os.path.join(self.path, 'db')

        if not os.path.exists(db_dir_path):
            os.mkdir(db_dir_path)

        return db_dir_path

    @property
    def db_file_path(self):
        return os.path.join(self.db_dir_path, DB_FILE_NAME)

    @property
    def db_version_path(self):
        return os.path.join(self.db_dir_path, DB_VERSION_NAME)

    @property
    def mitmproxy_options_json_path(self):
        return os.path.join(self.tmp_dir_path, MITMPROXY_OPTIONS_FILE_NAME)

    @property
    def ca_certs_dir_path(self):
        path = os.path.join(self.path, 'ca_certs')

        if not os.path.exists(path):
            os.makedirs(path)

        return path

    @property
    def settings_file_path(self):
        return os.path.join(self.path, 'settings.yml')

    @property
    def snapshots_dir_path(self):
        snapshots_dir_path = os.path.join(self.path, 'snapshots')

        if not os.path.exists(snapshots_dir_path):
            os.mkdir(snapshots_dir_path)

        return snapshots_dir_path

    @property
    def snapshots_history_dir_path(self):
        snapshots_history_dir_path = os.path.join(self.snapshots_dir_path, 'history')

        if not os.path.exists(snapshots_history_dir_path):
            os.mkdir(snapshots_history_dir_path)

        return snapshots_history_dir_path

    @property
    def snapshots_log_file_path(self):
        snapshots_log_file_path = os.path.join(self.snapshots_dir_path, 'log')

        return snapshots_log_file_path

    @property
    def snapshots_requests_dir_path(self):
        base_path = self.snapshots_dir_path
        requests_dir_path = os.path.join(base_path, 'requests')

        if not os.path.exists(requests_dir_path):
            os.mkdir(requests_dir_path)

        return requests_dir_path

    @property
    def snapshots_scenarios_dir_path(self):
        base_path = self.snapshots_dir_path
        scenarios_dir_path = os.path.join(base_path, 'scenarios')

        if not os.path.exists(scenarios_dir_path):
            os.mkdir(scenarios_dir_path)

        return scenarios_dir_path

    @property
    def snapshots_scenario_requests_dir_path(self):
        base_path = self.snapshots_scenarios_dir_path
        requests_dir_path = os.path.join(base_path, 'requests')

        if not os.path.exists(requests_dir_path):
            os.mkdir(requests_dir_path)

        return requests_dir_path

    @property
    def snapshosts_version_path(self):
        return os.path.join(self.snapshots_dir_path, 'VERSION')

    def remove(self, directory_path = None):
        if directory_path:
            data_dir_path = os.path.join(directory_path, DATA_DIR_NAME)
        else:
            data_dir_path = self.path

        if os.path.exists(data_dir_path):
            shutil.rmtree(data_dir_path)

    def create(self, directory_path = None):
        if not directory_path:
            directory_path = os.getcwd()

        self.__data_dir_path = os.path.join(directory_path, DATA_DIR_NAME)

        if not os.path.exists(self.__data_dir_path):
            os.mkdir(self.__data_dir_path)

            # Create the certs_dir_path if it doesn't exist
            self.certs_dir_path
            # Create tmp folder
            os.makedirs(os.path.join(self.__data_dir_path, TMP_DIR_NAME), exist_ok=True)

            with open(os.path.join(self.__data_dir_path, '.gitignore'), 'w') as fp:
                fp.write(
                    "\n".join([
                        'ca_certs',
                        'certs',
                        'db',
                        'settings.yml',
                        os.path.join('snapshots', 'log'),
                        os.path.join('snapshots', 'VERSION'),
                        TMP_DIR_NAME
                    ])
                )

    # If the current working directory does not contain a .stoobly folder,
    # then search in the parent directories until the home directory.
    def find_data_dir(self, start_path: str) -> str:
        root_dir = os.path.abspath(os.sep)

        while True:
            data_dir_path = os.path.join(start_path, DATA_DIR_NAME)
            if os.path.exists(data_dir_path) and os.stat(data_dir_path).st_uid == os.getuid():
                return data_dir_path

            # Root reached
            if start_path == root_dir:
                break

            # Move up one directory
            start_path = os.path.dirname(start_path)

        return os.path.join(os.path.expanduser('~'), DATA_DIR_NAME)



import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.cli import record, scenario
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

@pytest.fixture(scope='class', autouse=True)
def settings():
  return reset()

class TestRequest():
  @pytest.fixture(scope='class')
  def created_scenario(self, runner: CliRunner):
      create_result = runner.invoke(scenario, ['create', 'test'])
      assert create_result.exit_code == 0
      return Scenario.last()

  @pytest.fixture(scope='class', autouse=True)
  def recorded_request_one(self, runner: CliRunner, created_scenario: Scenario):
      record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      return Request.last()

  @pytest.fixture(scope='class', autouse=True)
  def recorded_request_two(self, runner: CliRunner, created_scenario: Scenario):
      record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), NON_DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      return Request.last()

  @pytest.fixture(scope='class', autouse=True)
  def snapshot_result(self, runner: CliRunner, created_scenario: Scenario):
    snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
    assert snapshot_result.exit_code == 0
    return snapshot_result

  def test_it_resets(self, runner: CliRunner, created_scenario: Scenario):
    requests = created_scenario.requests
    assert len(requests) == 2

    requests[1].delete()

    snapshot_result = runner.invoke(scenario, ['reset', created_scenario.key()])
    assert snapshot_result.exit_code == 0

    requests = Scenario.find(created_scenario.id).requests
    assert len(requests) == 2


import pdb

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.app.models.replayed_response_model import ReplayedResponseModel
from stoobly_agent.app.settings import Settings

class ReplayedResponseHeadersController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # GET /requests/:requestId/replayed_response/:replayedResponseId/headers
    def index(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'request_id': 1,
            'replayed_response_id': 3,
        })

        replayed_response_model = self.__replayed_response_model(context)
        raw, status = replayed_response_model.raw(context.params.get('replayed_response_id'))

        if context.filter_response(raw, status):
            return

        response = RawHttpResponseAdapter(raw).to_response()

        headers = []
        for key, val in response.headers.items():
            headers.append({
                'name': key,
                'value': val,
            })

        context.render(
            json = headers,
            status = 200
        )

    def __replayed_response_model(self, context: SimpleHTTPRequestHandler):
        access_token = context.headers.get('access-token')
        replayed_response_model = ReplayedResponseModel(Settings.instance(), access_token=access_token)
        return replayed_response_model


import requests

from typing import Union

from .api import Api

class AgentApi(Api):
    STATUSES_ENDPOINT = '/statuses'

    def __init__(self, service_url):
        self.service_url = service_url

    @property
    def default_headers(self):
        return {
            'X-Do-Proxy': '1',
        }

    def update_status(self, status_id: str, project_id: Union[int, str]):
        url = f"{self.service_url}{self.STATUSES_ENDPOINT}/{status_id}"
        return self.put(url, data=str(project_id), headers=self.default_headers)



# This file defines lifecyle hooks for when Stoobly intercepts a request
from stoobly_agent.app.proxy.record.context import RecordContext

def handle_before_request(context: RecordContext):
    intercept_settings = context.intercept_settings
    flow = context.flow
    request = flow.request

    # For example, uncomment the following
    #print(f"Agent running in {intercept_settings.mode} mode with {intercept_settings.policy} policy")
    #print(dir(request))
        


import pdb
import pytest

from typing import List

from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log, LogEvent
from stoobly_agent.app.models.factories.resource.local_db.helpers.log_event import PUT_ACTION, LogEvent, REQUEST_RESOURCE
from stoobly_agent.test.test_helper import reset

class TestLog():
  @pytest.fixture(scope='class', autouse=True)
  def settings(self):
    return reset()

  class TestRemovingProcessedEvents():

    @pytest.fixture(scope='class')
    def events(self):
      return [
        LogEvent(LogEvent.serialize(REQUEST_RESOURCE, '1', PUT_ACTION)),
        LogEvent(LogEvent.serialize(REQUEST_RESOURCE, '2', PUT_ACTION)),
        LogEvent(LogEvent.serialize(REQUEST_RESOURCE, '3', PUT_ACTION)),
        LogEvent(LogEvent.serialize(REQUEST_RESOURCE, '4', PUT_ACTION))
      ]

    @pytest.fixture(scope='class')
    def version_uuids(self, events: List[LogEvent]):
      log = Log()
      uuids = log.uuids(events)
      return log.generate_version(uuids)

    @pytest.fixture(scope='class')
    def branch_event(self):
      return LogEvent(LogEvent.serialize(REQUEST_RESOURCE, 'a', PUT_ACTION))

    class TestBranchPointAtStart():
      @pytest.fixture(scope='class')
      def log_events(self, events: List[LogEvent], branch_event: LogEvent):
        return [branch_event] + events

      def test_it_returns_event_at_start(self, log_events: List[LogEvent], version_uuids: List[str], branch_event: LogEvent):
        log = Log()
        unprocessed_events = log.remove_processed_events(log_events, version_uuids)
        assert log.uuids(unprocessed_events) == [branch_event.uuid]

    class TestBranchPointAtEnd():
      @pytest.fixture(scope='class')
      def log_events(self, events: List[LogEvent], branch_event: LogEvent):
        return events + [branch_event] 

      def test_it_returns_event_at_start(self, log_events: List[LogEvent], version_uuids: List[str], branch_event: LogEvent):
        log = Log()
        unprocessed_events = log.remove_processed_events(log_events, version_uuids)
        assert log.uuids(unprocessed_events) == [branch_event.uuid] 

    class TestBranchPointInMiddle():
      @pytest.fixture(scope='class')
      def log_events(self, events: List[LogEvent], branch_event: LogEvent):
        midpoint = int(len(events) / 2)
        return events[0:midpoint] + [branch_event] + events[midpoint:]

      def test_it_returns_event_in_middle(self, log_events: List[LogEvent], version_uuids: List[str], branch_event: LogEvent):
        log = Log()
        unprocessed_events = log.remove_processed_events(log_events, version_uuids)
        assert log.uuids(unprocessed_events) == [branch_event.uuid]

    class TestMultipleBranchEvents():
      @pytest.fixture(scope='class')
      def branch_events(self):
        return [
          LogEvent(LogEvent.serialize(REQUEST_RESOURCE, 'a', PUT_ACTION)),
          LogEvent(LogEvent.serialize(REQUEST_RESOURCE, 'b', PUT_ACTION))
        ]

      @pytest.fixture(scope='class')
      def log_events(self, events: List[LogEvent], branch_events: LogEvent):
        midpoint = int(len(events) / 2)
        return events[0:midpoint] + branch_events + events[midpoint:]

      def test_it_returns_branch_events(self, log_events: List[LogEvent], version_uuids: List[str], branch_events: LogEvent):
        log = Log()
        unprocessed_events = log.remove_processed_events(log_events, version_uuids)
        assert log.uuids(unprocessed_events) == log.uuids(branch_events)


import pdb

from .parameter_rule import ParameterRule
from .types.proxy_settings import RewriteRule as IRewriteRule
from .url_rule import UrlRule

class RewriteRule:

  def __init__(self, rewrite_rule: IRewriteRule):
    self.__rewrite_rule = rewrite_rule

    self.__methods = self.__rewrite_rule.get('methods') or []
    self.__pattern = self.__rewrite_rule.get('pattern')

    self.__raw_parameter_rules = self.__rewrite_rule.get('parameter_rules') or []
    self.__parameter_rules = list(map(lambda rule: ParameterRule(rule), self.__raw_parameter_rules))

    self.__raw_url_rules = self.__rewrite_rule.get('url_rules') or []
    self.__url_rules = list(map(lambda rule: UrlRule(rule), self.__raw_url_rules))

  @property 
  def methods(self):
    return self.__methods

  @property
  def pattern(self):
    return self.__pattern

  @property
  def parameter_rules(self):
    return self.__parameter_rules

  @parameter_rules.setter
  def parameter_rules(self, v):
    self.__parameter_rules = v

  @property
  def url_rules(self):
    return self.__url_rules

  @url_rules.setter
  def url_rules(self, v):
    self.__url_rules = v

  def to_dict(self):
    return {
      'methods': self.__methods,
      'pattern': self.__pattern,
      'parameter_rules': list(map(lambda parameter: parameter.to_dict(), self.__parameter_rules)),
      'url_rules': list(map(lambda url: url.to_dict(), self.__url_rules)),
    }


import json
import requests
import sys
from typing import Dict, TypedDict, Optional

from .endpoints_import_context import EndpointsImportContext
from stoobly_agent.app.models.types import ENDPOINT_COMPONENT_NAMES
from stoobly_agent.lib.api.body_param_names_resource import BodyParamNamesResource
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.header_names_resource import HeaderNamesResource
from stoobly_agent.lib.api.interfaces.endpoints import BodyParamName
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse
from stoobly_agent.lib.api.interfaces.endpoints import RequestComponentName
from stoobly_agent.lib.api.query_param_names_resource import QueryParamNamesResource
from stoobly_agent.lib.api.response_header_names_resource import ResponseHeaderNamesResource
from stoobly_agent.lib.api.response_param_names_resource import ResponseParamNamesResource
from stoobly_agent.lib.logger import bcolors

class ComponentImportArgs(TypedDict):
    project_id: int
    endpoint_id: int
    resource: EndpointsResource
    component: RequestComponentName
    parents: Optional[Dict]

def import_endpoints(context: EndpointsImportContext):
    for endpoint in context.endpoints:
        for endpoint_handler in context.endpoint_handlers:
            endpoint_handler(endpoint)

        try:    
            res = import_endpoint(context.project_id, context.resources['endpoint'], endpoint)
        except requests.HTTPError as e:
            error_handler(e, f"Failed to import endpoint: {endpoint['method']} {endpoint['path']}")
            return
        
        if res.status_code == 409: # Endpoint already created
            continue

        endpoint_id = res.json()['id']
        try:
            res = import_component_names(context.project_id, endpoint_id, endpoint, context.resources)
        except requests.HTTPError as e:
            error_handler(e, f"Failed to import endpoint: {endpoint['method']} {endpoint['path']}")

            try:
                cleanup_endpoint(context.project_id, endpoint_id, context.resources['endpoint'])
            except requests.HTTPError as e:
                error_handler(e, f"Failed to delete endpoint: {endpoint['method']} {endpoint['path']}, id {endpoint_id}")
            return

def import_endpoint(project_id: int, endpoint_resource: EndpointsResource, endpoint: EndpointShowResponse):
    res: requests.Response = endpoint_resource.create(**{
        'host': endpoint.get('host'),
        'method': endpoint.get('method'),
        'path_segments': json.dumps(endpoint.get('path_segment_names', [])),
        'path': endpoint.get('path'),
        'port': endpoint.get('port'),
        'project_id': project_id,
    })

    if res.ok or res.status_code == 409: # Endpoint already created
        return res
    
    res.raise_for_status()

def import_header_name(args: ComponentImportArgs):
    project_id = args['project_id']
    endpoint_id = args['endpoint_id']
    header_name_resource = args['resource']
    header_name = args['component']

    res: requests.Response = header_name_resource.create(endpoint_id, {
        'name': header_name.get('name'),
        'is_deterministic': header_name.get('is_deterministic', True),
        'is_required': header_name.get('is_required', False),
        'endpoint_id': endpoint_id,
        'project_id': project_id,  
    })

    res.raise_for_status()

def import_body_param_name(args: ComponentImportArgs):
    project_id = args['project_id']
    endpoint_id = args['endpoint_id']
    body_param_name_resource = args['resource']
    body_param_name = args['component']
    parents = args['parents']

    body_param_name_id = None
    parent_id = body_param_name.get('body_param_name_id')
    if parent_id:
        body_param_name_id = parents[parent_id]
    
    res: requests.Response = body_param_name_resource.create(endpoint_id, {
        'name': body_param_name.get('name'),
        'is_deterministic': body_param_name.get('is_deterministic', True),
        'is_required': body_param_name.get('is_required', False),
        'inferred_type': body_param_name.get('inferred_type'),
        'query': body_param_name.get('query'),
        'endpoint_id': endpoint_id,
        'project_id': project_id,
        'body_param_name_id': body_param_name_id
    })

    res.raise_for_status()

    res_id = res.json()['id']
    parents[body_param_name['id']] = res_id

def import_query_param_name(args: ComponentImportArgs):
    project_id = args['project_id']
    endpoint_id = args['endpoint_id']
    query_param_name_resource = args['resource']
    query_param_name = args['component']

    res: requests.Response = query_param_name_resource.create(endpoint_id, {
        'name': query_param_name.get('name'),
        'is_deterministic': query_param_name.get('is_deterministic', True),
        'is_required': query_param_name.get('is_required', False),
        'endpoint_id': endpoint_id,
        'project_id': project_id,
    })

    res.raise_for_status()

def import_response_param_name(args: ComponentImportArgs):
    project_id = args['project_id']
    endpoint_id = args['endpoint_id']
    response_param_name_resource = args['resource']
    response_param_name = args['component']
    parents = args['parents']

    response_param_name_id = None
    parent_id = response_param_name.get('response_param_name_id')
    if parent_id:
        response_param_name_id = parents[parent_id]
    
    res: requests.Response = response_param_name_resource.create(endpoint_id, {
        'name': response_param_name.get('name'),
        'is_deterministic': response_param_name.get('is_deterministic', True),
        'is_required': response_param_name.get('is_required', False),
        'inferred_type': response_param_name.get('inferred_type'),
        'query': response_param_name.get('query'),
        'endpoint_id': endpoint_id,
        'project_id': project_id,
        'response_param_name_id': response_param_name_id
    })

    res.raise_for_status()

    res_id = res.json()['id']
    parents[response_param_name['id']] = res_id


def import_response_header_name(args: ComponentImportArgs):
    project_id = args['project_id']
    endpoint_id = args['endpoint_id']
    response_header_name_resource = args['resource']
    response_header_name = args['component']

    res: requests.Response = response_header_name_resource.create(endpoint_id, {
        'name': response_header_name.get('name'),
        'is_deterministic': response_header_name.get('is_deterministic', True),
        'is_required': response_header_name.get('is_required', False),
        'endpoint_id': endpoint_id,
        'project_id': project_id,  
    })

    res.raise_for_status()

component_name_import_dispatch = {
    ENDPOINT_COMPONENT_NAMES[0]: import_header_name,
    ENDPOINT_COMPONENT_NAMES[1]: import_body_param_name,
    ENDPOINT_COMPONENT_NAMES[2]: import_query_param_name,
    ENDPOINT_COMPONENT_NAMES[3]: import_response_header_name,
    ENDPOINT_COMPONENT_NAMES[4]: import_response_param_name
}

def import_component_names(project_id: int, endpoint_id: int, endpoint: EndpointShowResponse, resources: Dict[str, EndpointsResource]):
    for component_name in ENDPOINT_COMPONENT_NAMES:
        parents: Dict[int, int] = {}
        for component in endpoint.get(f'{component_name}s', {}):
            resource = resources[component_name]
            args = {
                'project_id': project_id,
                'endpoint_id': endpoint_id,
                'resource': resource,
                'component': component,
                'parents': parents
            }
            component_name_import_dispatch[component_name](args)
            

def cleanup_endpoint(project_id: int, endpoint_id: int, resource: EndpointsResource):
    print(f"{bcolors.OKBLUE}Cleaning up partial import...{bcolors.ENDC}")
    res: requests.Response = resource.destroy(endpoint_id, **{
        'project_id': project_id
    })

    res.raise_for_status()

def error_handler(error: requests.HTTPError, message = ''):
    if message:
        print(
            f"{bcolors.FAIL}{message}{bcolors.ENDC}",
            file=sys.stderr
        )

    print(error.response.text, file=sys.stderr)
    print(f"Error {error.response.status_code} {error.response.reason}", file=sys.stderr)


import pdb
import requests

from stoobly_agent.lib.api.projects_resource import ProjectsResource
from stoobly_agent.lib.api.users_resource import UsersResource
from stoobly_agent.lib.api.interfaces import ProjectDetails, ProjectShowResponse, ProjectsIndexResponse, UserProfileResponse
from stoobly_agent.lib.api.keys import OrganizationKey, ProjectKey
from stoobly_agent.app.settings import Settings

class ProjectFacade():

  def __init__(self, __settings: Settings):
    self.__settings = __settings
    self.__api = ProjectsResource(self.__settings.remote.api_url, self.__settings.remote.api_key)

  def user_profile(self) -> UserProfileResponse:
    resource = UsersResource(self.__settings.remote.api_url, self.__settings.remote.api_key)
    res = resource.profile()

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()

  def create(self, **kwargs: ProjectDetails) -> ProjectShowResponse:
    organization_key: OrganizationKey = OrganizationKey(kwargs.get('organization_key')) 

    res: requests.Response =  self.__api.create(
      project={
        "description": kwargs.get('description') or '',
        "name": kwargs.get('name'),
        "organization_id": organization_key.id
      },
      organization_id=organization_key.id
    )

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()

  def show(self, project_key: str) -> ProjectShowResponse:
    key = ProjectKey(project_key)
    res = self.__api.show(key.id)

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()

  def index(self, kwargs) -> ProjectsIndexResponse:
    organization_key: str = OrganizationKey(kwargs.get('organization_key')) 

    del kwargs['organization_key']

    res = self.__api.index(**{ 'organization_id': organization_key.id,  **kwargs })

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()




import pdb
import requests

from stoobly_agent.app.models.types.request_components import Header
from stoobly_agent.app.settings import Settings

from .factories.resource.response_header import ResponseHeaderResourceFactory
from .model import Model

class ResponseHeaderModel(Model):

  def __init__(self, settings: Settings, **options):
    super().__init__(settings, **options)

  def as_local(self):
    self.adapter =  ResponseHeaderResourceFactory(self.settings.remote).local_db()
    self.is_local = True

  def as_remote(self):
    self.is_local = False

  def create(self, request_id: int, **params: Header):
    try:
      return self.adapter.create(request_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def index(self, request_id: str, **query_params):
    try:
      return self.adapter.index(request_id, **query_params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def update(self, request_id: int, **params: Header):
    try:
      return self.adapter.update(request_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def destroy(self, request_id: str, id: str):
    try:
      return self.adapter.destroy(request_id, id)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)


import os
import shutil

from stoobly_agent.config.data_dir import DataDir

class Snapshot():

  def __init__(self, uuid: str):
    self.__uuid = uuid
    self.__data_dir = DataDir.instance()

  @property
  def uuid(self):
    return self.__uuid

  @property
  def data_dir(self):
    return self.__data_dir

  def copy_file(self, src: str, dest_dir: str):
    if not os.path.exists(src):
      return None

    data_dir_parent = os.path.dirname(self.data_dir.path)
    dest_file_path = src.replace(data_dir_parent, dest_dir)
    dest_dir_path = os.path.dirname(dest_file_path)

    if not os.path.exists(dest_dir_path):
      os.makedirs(dest_dir_path, exist_ok=True)

    shutil.copy(src, dest_file_path)

    return dest_file_path


import os
import pdb

from stoobly_agent.config.constants import env_vars

from .data_rules import DataRules
from .types.proxy_settings import DataSettings as IDataSettings

class DataSettings:
  __data_settings = None

  def __init__(self, data_settings: IDataSettings):
    self.__data_settings = data_settings or {}
    self.__data_rules_map = {k: DataRules(v) for k, v in self.__data_settings.items()}

  def data_rules(self, project_id: str) -> DataRules:
    if not self.__data_rules_map.get(project_id):
      self.__data_rules_map[project_id] = DataRules({})
    return self.__data_rules_map[project_id]

  def to_dict(self):
    return {k: v.to_dict() for k, v in self.__data_rules_map.items()}


import os
import pathlib
import shutil

from .app import App
from .app_config import AppConfig
from .command import Command
from .docker.constants import DOCKER_COMPOSE_NETWORKS

class AppCommand(Command):

  def __init__(self, app: App):
    super().__init__(app)

    self.__config = AppConfig(self.scaffold_namespace_path)

  @property
  def app_dir_path(self):
    return self.app.dir_path

  @property
  def app_dir_exists(self):
    return os.path.exists(self.app_dir_path)

  @property
  def app_config(self):
    return self.__config

  @property
  def app_config_path(self):
    return self.__config.path

  @property
  def networks_compose_path(self):
    return os.path.join(
      self.data_dir_path,
      self.networks_compose_relative_path
    )

  @property
  def networks_compose_relative_path(self):
    return os.path.join(
      self.namespace,
      DOCKER_COMPOSE_NETWORKS
    )

  @property
  def scaffold_namespace_path(self):
    return self.app.scaffold_namespace_path

  @property
  def app_templates_root_dir(self):
    return os.path.join(self.templates_root_dir, 'app')

  @property
  def data_dir_path(self):
    return self.app.data_dir_path

  @property
  def scaffold_namespace_path(self):
    return self.app.scaffold_namespace_path

  @property
  def templates_root_dir(self):
    return os.path.join(pathlib.Path(__file__).parent.resolve(), 'templates')

  def config(self, _c: dict):
    _config = self.app_config.read()
    _config.update(_c)
    return _config

  def copy_files_no_replace(self, source_dir: str, src_files: list, dest_dir: str):
    return self.copy_files(source_dir, src_files, dest_dir, False)

  def copy_files(self, source_dir: str, src_files: list, dest_dir: str, replace_ok=True):
    # Ensure the destination directory exists
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    
    for file in src_files:
      src_file_path = os.path.join(source_dir, file)
      if not os.path.isfile(src_file_path):
        continue
      
      if not os.path.exists(src_file_path):
        continue

      dest_subdir = os.path.join(dest_dir, os.path.dirname(file))

      if os.path.exists(os.path.join(dest_dir, file)) and not replace_ok:
        continue

      if not os.path.exists(dest_subdir):
        os.makedirs(dest_subdir, exist_ok=True)

      shutil.copy(src_file_path, dest_subdir)



import pdb
import pytest

from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.app.proxy.test.matchers.context import MatchContext
from stoobly_agent.app.proxy.test.matchers.diff import list_matches
from stoobly_agent.config.constants import test_filter
from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias
from stoobly_agent.test.mock_data.endpoint_show_response import endpoint_show_response

@pytest.fixture
def endpoints_list_response_param_names_facade():
  response_param_names = endpoint_show_response['response_param_names']
  aliases = endpoint_show_response['aliases']
  return RequestComponentNamesFacade(response_param_names).with_aliases(aliases)

@pytest.fixture
def empty_response_param_names_facade():
  response_param_names = []
  return RequestComponentNamesFacade(response_param_names)

class TestMatchesDiffSanity():
  def test_matches_empty_response(self, endpoints_list_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      [], [], 
    )
    assert matches, log
  
  def test_matches_list_of_numbers(self, empty_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': empty_response_param_names_facade
      }),
      [1], [1]
    )
    assert matches, log

  def test_not_matches(self, endpoints_list_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      [], {}
    )
    assert log == "Key '' type did not match: got <class 'dict'>, expected <class 'list'>", log

  def test_not_matches_list_of_numbers(self, empty_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': empty_response_param_names_facade
      }),
      [1], [2]
    )

    assert log == "Key '[0]' did not match: got 2, expected 1", log

  def test_not_matches_different_length(self, empty_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': empty_response_param_names_facade
      }),
      [1], [1, 2]
    )

    assert log == "Key '' length did not match: got 2, expected 1", log

class TestMatchesListOfDicts():
  def test_matches_id_not_deterministic(self, endpoints_list_response_param_names_facade):
    actual = [
      {
        "id": 1000,
        "requests_count": 6,
        "category": 1,
        "path": "/abc",
        "method": "GET",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ]

    expected = [
      {
        "id": 930,
        "requests_count": 6,
        "category": 1,
        "path": "/abc",
        "method": "GET",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ] 

    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      expected, actual
    )
    assert matches, log

  def test_not_matches_requests_count(self, endpoints_list_response_param_names_facade):
    actual = [
      {
        "id": 1000,
        "requests_count": 8,
        "category": 2,
        "path": "/abc",
        "method": "POST",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ]

    expected = [
      {
        "id": 1000,
        "requests_count": 6,
        "category": 2,
        "path": "/abc",
        "method": "POST",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ] 

    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      expected, actual
    )

    assert log == "Key '[0].requests_count' did not match: got 8, expected 6", log

class TestAliasFilterMatchesListOfDicts():
  def test_matches_id_aliased(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    endpoints_list_response_param_names_facade.with_filter(test_filter.ALIAS)

    actual = [
      {
        "id": 1000,
        "requests_count": 6,
        "category": 1,
        "path": "/abc",
        "method": "GET",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ]

    expected = [
      {
        "id": 1000,
        "requests_count": 6,
        "category": 2,
        "path": "/abcdef",
        "method": "POST",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": None,
        "url": "https://alpha.api.stoobly.com/endpoints/931"
      },
    ] 

    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      expected, actual
    )
    assert matches, log

  def test_not_matches_id_aliased(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    endpoints_list_response_param_names_facade.with_filter(test_filter.ALIAS)

    # Set aliased params as deterministic to prevent ignore due to not deterministic
    for response_param_name in endpoints_list_response_param_names_facade.aliased:
      response_param_name['is_deterministic'] = True

    actual = [
      {
        "id": 1000,
        "requests_count": 6,
        "category": 1,
        "path": "/abc",
        "method": "GET",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ]

    expected = [
      {
        "id": 1001,
        "requests_count": 6,
        "category": 2,
        "path": "/abcdef",
        "method": "POST",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": None,
        "url": "https://alpha.api.stoobly.com/endpoints/931"
      },
    ] 

    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      expected, actual
    )

    assert log == "Key '[0].id' did not match: got 1000, expected 1001", log

  class TestLinkFilterMatchesListOfDicts():
    def test_matches_id_aliased(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
      trace = Trace.create()

      TraceAlias.create(trace_id=trace.id, name=':endpointId', value=1001)
      TraceAlias.create(trace_id=trace.id, name=':endpointRequestCounts', value=2)

      endpoints_list_response_param_names_facade.with_filter(test_filter.LINK).with_trace(trace)

      actual = [
        {
          "id": 1000, # Aliased as :endpointId
          "requests_count": 7, # Aliased as :endpointRequestsCount
          "category": 1,
          "path": "/abc",
          "method": "GET",
          "created_at": "2022-05-26T23:46:49.968Z",
          "updated_at": "2022-05-26T23:46:49.968Z",
          "components": [],
          "url": "https://alpha.api.stoobly.com/endpoints/930"
        },
      ]

      expected = [
        {
          "id": 1000,
          "requests_count": 7,
          "category": 2,
          "path": "/abcdef",
          "method": "POST",
          "created_at": "2022-05-26T23:46:49.968Z",
          "updated_at": "2022-05-26T23:46:49.968Z",
          "components": None,
          "url": "https://alpha.api.stoobly.com/endpoints/931"
        },
      ] 

      matches, log = list_matches(
        MatchContext({
          'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
        }),
        expected, actual
      )
      assert matches, log

    def test_not_matches_id_aliased(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
      trace = Trace.create()

      TraceAlias.create(trace_id=trace.id, name=':endpointId', value=1001)

      endpoints_list_response_param_names_facade.with_filter(test_filter.LINK).with_trace(trace)

      actual = [
        {
          "id": 1001, # Aliased as :endpointId
          "requests_count": 8, # Aliased as :endpointRequestsCount
          "category": 1,
          "path": "/abc",
          "method": "GET",
          "created_at": "2022-05-26T23:46:49.968Z",
          "updated_at": "2022-05-26T23:46:49.968Z",
          "components": [],
          "url": "https://alpha.api.stoobly.com/endpoints/930"
        },
      ]

      expected = [
        {
          "id": 1000,
          "requests_count": 7,
          "category": 2,
          "path": "/abcdef",
          "method": "POST",
          "created_at": "2022-05-26T23:46:49.968Z",
          "updated_at": "2022-05-26T23:46:49.968Z",
          "components": None,
          "url": "https://alpha.api.stoobly.com/endpoints/931"
        },
      ] 

      matches, log = list_matches(
        MatchContext({
          'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
        }), 
        expected, actual, 
      )

      assert log == "Key '[0].id' did not match: got 1001, expected 1000", log


import json
import pdb
import requests
import sys

from typing import List, TypedDict, Union

from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.cli.types.output import ReplayOutput
from stoobly_agent.app.proxy.replay.body_parser_service import decode_response, is_traversable
from stoobly_agent.lib.logger import bcolors
from stoobly_agent.lib.utils import jmespath
from stoobly_agent.lib.utils.decode import decode

BODY_FORMAT = 'body'
DEFAULT_FORMAT = 'default'
JSON_FORMAT = 'json'

class ReplaySession(TypedDict):
  buffer: bool
  contexts: List[ReplayContext] 
  format: Union[BODY_FORMAT, DEFAULT_FORMAT, JSON_FORMAT, None]
  scenario_id: int
  total: int

def handle_before_replay(context: ReplayContext, session: ReplaySession):
  format = session.get('format') or DEFAULT_FORMAT

  if format == DEFAULT_FORMAT:
    request = context.request

    # If request is not the first of a sequence, print extra new line
    sequence = context.sequence
    if sequence and sequence > 1:
      print()

    print(f"{bcolors.OKCYAN}{request.method} {request.url}{bcolors.ENDC}")

def handle_after_replay(context: ReplayContext, session: ReplaySession):
  if not 'contexts' in session:
    session['contexts'] = []

  if not 'total' in session:
    session['total'] = 0

  if not session.get('buffer'):
    print_with_decoding(format_request(context))
  else:
    session['contexts'].append(context)

  session['total'] += 1

def format_request(context: ReplayContext, format = None):
  format_handler = __default_format_handler 
  
  if format == JSON_FORMAT:
    format_handler = __json_format_handler
  elif format == BODY_FORMAT:
    format_handler = __body_format_handler

  return format_handler(context)

def print_request_query(context: ReplayContext, query: str):
  response = context.response
  content = __content(response)
  content_type = response.headers.get('content-type')

  decoded_response = decode_response(content, content_type)
  if not is_traversable(decoded_response):
    print(
      f"{bcolors.FAIL}Could not query request, expected responsed to be of type {dict} or {list}, got {decoded_response.__class__} {bcolors.ENDC}",
      file=sys.stderr
    )
    print("\n".join(format_request(context)), file=sys.stderr)
  else:
    print(jmespath.search(query, decoded_response))

def print_session(session: ReplaySession):
  if len(session['contexts']) == 0:
    return 

  format = session.get('format') or DEFAULT_FORMAT

  outputs = []

  for replay_context in session['contexts']:
    outputs += format_request(replay_context, format)

  if format == JSON_FORMAT:
    if session.get('scenario_id'):
      print(json.dumps(outputs))
    else:
      print(json.dumps(outputs[0]))
  else:
    print_with_decoding(outputs)

def print_with_decoding(outputs: List[Union[bytes, str]], delimitter = "\n"):
  for output in outputs:
    o = decode(output)

    if isinstance(o, bytes):
      sys.stdout.buffer.write(o + delimitter.encode())
    else:
      print(o, end=delimitter)

def __default_format_handler(context: ReplayContext, additional=''):
  output = [context.response_content]

  response = context.response
  seconds = context.end_time - context.start_time
  ms = round(seconds * 1000)
  output.append(f"Completed {response.status_code} in {ms}ms{additional}")

  return output

def __body_format_handler(context: ReplayContext):
  return [context.response_content]

def __json_format_handler(context: ReplayContext):
  request = context.request
  response = context.response

  content = __content(response)
  headers = dict(response.headers)
  method = request.method
  url = request.url

  seconds = context.end_time - context.start_time
  ms = round(seconds * 1000)
  output: ReplayOutput = {'content': content, 'headers': headers, 'latency': ms, 'method': method, 'url': url}

  return [output]

def __content(res: requests.Response):
  content = res.content
  return decode(content)


import pdb
import pytest

from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.app.proxy.test.matchers.context import MatchContext
from stoobly_agent.app.proxy.test.matchers.contract import list_matches
from stoobly_agent.config.constants.lifecycle_hooks import ON_PARAM_NAME_MISSING_ERROR
from stoobly_agent.test.mock_data.endpoint_show_response import endpoint_show_response

@pytest.fixture
def endpoints_list_response_param_names_facade():
  response_param_names = endpoint_show_response['response_param_names']
  aliases = endpoint_show_response['aliases']
  return RequestComponentNamesFacade(response_param_names).with_aliases(aliases)

class TestListMatchesList():

  def test_matches(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      [{
        "id": 934,
        "requests_count": 2,
        "category": 2,
        "path": "/users/:userId",
        "method": "GET",
        "created_at": "2022-05-31T06:00:10.663Z",
        "updated_at": "2022-05-31T06:00:10.663Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/934"
      }]
    )
    assert matches, log

  def test_not_matches_list_type(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      [1], strict=True
    )

    assert not matches
    assert log == "Key '[0]' type did not match: got <class 'int'>, expected valid types Hash", log

  def test_not_matches_list_element_type(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    lifecycle_hooks = {}
    lifecycle_hooks[ON_PARAM_NAME_MISSING_ERROR] = lambda context, actual: True # Ignore this error

    match_context = MatchContext({
      'lifecycle_hooks': lifecycle_hooks, 'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade 
    })

    matches, log = list_matches(
      match_context, [{ "id": "934" }], strict=True
    )

    assert not matches
    assert log == "Key '[0].id' type did not match: got <class 'str'>, expected valid types Integer", log

  def test_not_matches_list_element_missing_properties(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):  
    match_context = MatchContext({
      'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
    })

    matches, log = list_matches(
      match_context, [{ "id": 934 }], strict=True
    )

    assert not matches
    assert log == "Missing key: expected [0].requests_count to exist", log

  def test_not_matches_list_element_properties(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    lifecycle_hooks = {}
    lifecycle_hooks[ON_PARAM_NAME_MISSING_ERROR] = lambda context, actual: True # Ignore this error
    match_context = MatchContext({
      'lifecycle_hooks': lifecycle_hooks, 'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
    })

    matches, log = list_matches(
      match_context, [{ "random_property": 934 }], strict=True
    )

    assert not matches
    assert log == "Extra key: expected [0].random_property to not exist", log

  def test_not_matches_list_length(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    match_context = MatchContext({
      'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
    })

    matches, log = list_matches(
      match_context, [], strict=True
    )

    assert not matches
    assert log == "Key '[*]' length did not match: got 0", log

  def test_not_matches_list(self, endpoints_list_response_param_names_facade: RequestComponentNamesFacade):
    match_context = MatchContext({
      'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
    })

    matches, log = list_matches(
      match_context, {}, strict=True
    )
    
    assert not matches
    assert log == "Key '[*]' type did not match: got <class 'dict'>, expected <class 'list'>", log


from .organization_key import InvalidOrganizationKey, OrganizationKey
from .project_key import InvalidProjectKey, ProjectKey
from .report_key import InvalidReportKey, ReportKey
from .request_key import InvalidRequestKey, RequestKey
from .scenario_key import InvalidScenarioKey, ScenarioKey
from .test_key import InvalidTestKey, TestKey


import pdb
import requests

from stoobly_agent.app.models.adapters.python import PythonRequestAdapterFactory
from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter

class TestRawAdapter():

  def test_www_google_com(self):
    request = requests.Request(
      data=b'',
      headers={
        'Content-Length': '0'
      },
      method='GET',
      url='https://www.google.com?a=1&a=2',
    )
    raw_request = PythonRequestAdapterFactory(request).raw_request()

    self.__test_equivalence(raw_request, request)

  def test_stoobly_com(self):
    data = b'{"name": "test", "description": "test"}'
    request = requests.Request(
      data=data,
      headers={
        'Content-Length': str(len(data)),
        'Content-Type': 'application/json',
      },
      method='POST',
      url='https://stoobly.com/scenarios',
    )
    raw_request = PythonRequestAdapterFactory(request).raw_request()

    self.__test_equivalence(raw_request, request)

  def __test_equivalence(self, raw_request: str, request: requests.Request):
    _request = RawHttpRequestAdapter(raw_request).to_request()

    assert _request.method == request.method
    assert _request.url == request.url
    assert _request.data == request.data

    for key, val in request.headers.items():
      assert val == _request.headers.get(key)




import pdb

from stoobly_agent.lib.cache import Cache
from stoobly_agent.lib.utils.decode import decode

class StatusesController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    def index(self, context):
        version = -1

        try:
            version = int(context.params.get('version'))
        except ValueError as e:
            pass

        cache = Cache.instance()
        statuses = []
        if not version or version != cache.version:
            statuses = cache.read_all()

        context.render(
            json = {
                'statuses': statuses,
                'version': cache.version,
            },
            status = 200
        )

    # GET /statuses/:id
    def get(self, context):
        context.parse_path_params({
            'id': 4
        })

        cache = Cache.instance()
        status = cache.read(context.params.get('id'))

        if not status:
            context.render(
                plain = '',
                status = 204
            )
        else:
            context.render(
                json = {
                    'status': status,
                    'version': cache.version
                },
                status = 200
            )

    # PUT /statuses/:id
    def update(self, context):
        context.parse_path_params({
            'id': 2
        })

        cache = Cache.instance()
        cache.write(context.params.get('id'), decode(context.body))

        context.render(
            plain = '',
            status = 200
        )




from io import BytesIO
from requests import Response

class JSONResponseBuilder():

  def __init__(self):
    self.__response = Response()
    self.__response.headers['Content-Type'] = 'application/json'

  def with_body(self, body: bytes):
    self.__response.raw = BytesIO(body)

  def with_status_code(self, status_code: int):
    self.__response.status_code = status_code


from typing import List, Optional, TypedDict, Union

from .pagination_query_params import PaginationQueryParams

class EndpointsIndexQueryParams(PaginationQueryParams):
  ignored_components: str
  method: str
  project_id: str
  q: str

class RequestComponentName(TypedDict):
  id: int 
  alias_id: int
  is_deterministic: bool
  is_required: bool
  name: str
  values: List[str] # Sample values for the component

class Alias(TypedDict):
  id: int
  name: str

class BodyParamName(RequestComponentName):
  inferred_type: str
  query: str
  body_param_name_id: Union[int, None]

class ResponseParamName(RequestComponentName):
  inferred_type: str
  query: str

class EndpointShowResponse(TypedDict):
  id: int
  method: str
  host: str
  ignored_components: Optional[list]
  port: str
  path: str
  match_pattern: Optional[str]
  aliases: List[Alias]
  body_param_names: List[BodyParamName]
  header_names: List[RequestComponentName]
  path_segment_names: List[RequestComponentName]
  query_param_names: List[RequestComponentName]
  response_param_names: List[ResponseParamName]
  response_header_names: List[RequestComponentName]
  literal_query_params: Optional[dict]
  literal_body_params: Optional[dict]
  literal_response_params: Optional[dict]

ARRAY_TYPE = 'Array'




import pdb

from typing import Dict, List, Union

from stoobly_agent.app.proxy.replay.alias_context import AliasContext
from stoobly_agent.app.proxy.replay.alias_resolver import AliasResolver
from stoobly_agent.config.constants import alias_resolve_strategy
from stoobly_agent.lib.api.interfaces.endpoints import Alias, RequestComponentName
from stoobly_agent.lib.orm.trace_alias import TraceAlias
from stoobly_agent.lib.utils import jmespath

AliasMap = Dict[str, RequestComponentName]

def __handle_after_replace(name, value, trace_alias: TraceAlias):
  pass

def __handle_replace(name, value, alias_context: AliasContext):
  trace_alias = alias_context.resolve_alias(name, value)

  if not trace_alias:
    raise ValueError('Alias not set')
  else:
    return trace_alias.value

def build_id_to_alias_map(aliases: List[Alias]):
    id_to_alias = {}
    for _alias in aliases:
      id_to_alias[_alias['id']] = _alias
    return id_to_alias

def rewrite_params(
  params: Union[list, dict],
  param_names: List[RequestComponentName], 
  id_to_alias: AliasMap, 
  alias_resolver: AliasResolver,
  **options: jmespath.Options
) -> None:
  # If no resolve strategy is set, return
  if alias_resolver.strategy == alias_resolve_strategy.NONE:
    return

  for param_name in param_names:
    alias_context = AliasContext(alias_resolver)
    _alias: Alias = id_to_alias.get(param_name['alias_id'])

    if not _alias:
      continue
 
    handle_after_replace = options.get('handle_after_replace')
    handle_after_replace = handle_after_replace if handle_after_replace else __handle_after_replace

    # We have may have to first search for all values matching query,
    # If there's more than one, then try to assign different alias values 
    handle_replace = options.get('handle_replace')
    handle_replace = handle_replace if handle_replace else __handle_replace

    query = param_name['query'] if 'query' in param_name else param_name['name']
    jmespath.search(
      query, params, { 
        'handle_replace': lambda name, value, i: handle_replace(_alias['name'], value, alias_context),
        'handle_after_replace': lambda name, value, i: handle_after_replace(name, value, alias_context.resolved_aliases[i])
      }
    )


import os
import pdb
import pytest
import time

from click.testing import CliRunner
from typing import List

from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.app.models.factories.resource.local_db.helpers.log_event import DELETE_ACTION, LogEvent
from stoobly_agent.app.models.factories.resource.local_db.helpers.scenario_snapshot import ScenarioSnapshot
from stoobly_agent.cli import record, request, scenario, snapshot
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.test.test_helper import assert_orm_request_equivalent, DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL, reset

def write_junk(path: str):
  with open(path, 'w') as fp:
    fp.write('junk')

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestApply():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
      return reset()

    class TestWhenDeletingRequest():

      @pytest.fixture(scope='class')
      def log(self):
        return Log()

      @pytest.fixture(scope='class')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(autouse=True, scope='class')
      def events(self, runner: CliRunner, recorded_request: Request, log: Log):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        events = log.events
        event: LogEvent = events[len(events) - 1]
        log.version = event.uuid

        time.sleep(1)

        snapshot_result = runner.invoke(request, ['snapshot', '--action', DELETE_ACTION, recorded_request.key()])
        assert snapshot_result.exit_code == 0

        return log.events

      @pytest.fixture(scope='class')
      def delete_event(self, events: List[LogEvent]):
        return events[1]

      def test_it_deletes(self, runner: CliRunner, recorded_request: Request, delete_event: LogEvent):
        apply_result = runner.invoke(snapshot, ['apply', '--force', delete_event.uuid])
        assert apply_result.exit_code == 0

        _request = Request.find(recorded_request.id)
        assert _request == None

    class TestWhenCreatingRequest():

      @pytest.fixture(scope='function')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(autouse=True, scope='function')
      def put_event(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        return events[len(events) - 1]

      def test_it_creates_request(self, runner: CliRunner, recorded_request: Request):
        recorded_request.delete()
        assert Request.find(recorded_request.id) == None

        apply_result = runner.invoke(snapshot, ['apply'])
        assert apply_result.exit_code == 0

        _request = Request.find_by(uuid=recorded_request.uuid)
        assert _request != None
        assert_orm_request_equivalent(_request, recorded_request)

      def test_idempotent(self, runner: CliRunner, recorded_request: Request, put_event: LogEvent):
        apply_result = runner.invoke(snapshot, ['apply', put_event.uuid])
        assert apply_result.exit_code == 0

        _request = Request.find_by(uuid=recorded_request.uuid)
        assert_orm_request_equivalent(_request, recorded_request)

    class TestWhenDeletingScenario():

      @pytest.fixture(scope='class')
      def log(self):
        return Log()

      @pytest.fixture(scope='class')
      def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

      @pytest.fixture(autouse=True, scope='class')
      def events(self, runner: CliRunner, created_scenario: Scenario, log: Log):
        snapshot_result = runner.invoke(scenario, ['snapshot',  created_scenario.key()])
        assert snapshot_result.exit_code == 0

        events = log.events
        event: LogEvent = events[len(events) - 1]
        log.version = event.uuid

        time.sleep(1)

        snapshot_result = runner.invoke(scenario, ['snapshot', '--action', DELETE_ACTION, created_scenario.key()])
        assert snapshot_result.exit_code == 0

        return log.events

      @pytest.fixture(scope='class')
      def delete_event(self, events):
        return events[1]

      def test_it_deletes(self, runner: CliRunner, created_scenario: Scenario, delete_event: LogEvent):
        snapshot_result = runner.invoke(snapshot, ['apply', '--force', delete_event.uuid])
        assert snapshot_result.exit_code == 0

        scenario = Scenario.find(created_scenario.id)

        assert scenario == None

    class TestWhenCreatingScenario():

      @pytest.fixture(scope='class')
      def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

      @pytest.fixture(scope='class', autouse=True)
      def created_scenario_requests(self, runner: CliRunner, created_scenario: Scenario):
          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), NON_DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          return created_scenario.requests

      @pytest.fixture(scope='class', autouse=True)
      def put_event(self, runner: CliRunner, created_scenario: Scenario):
        snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        return events[len(events) - 1]

      @pytest.fixture(scope='class', autouse=True)
      def recreated_scenario(self, runner: CliRunner, created_scenario: Scenario):
        requests = created_scenario.requests

        created_scenario.delete()
        assert Scenario.find(created_scenario.id) == None

        for request in requests:
          request.delete()
          assert Request.find(request.id) == None

        apply_result = runner.invoke(snapshot, ['apply'])
        assert apply_result.exit_code == 0

        _scenario = Scenario.find_by(uuid=created_scenario.uuid)
        assert _scenario != None

        return _scenario

      def test_it_creates_scenario(self, created_scenario: Scenario, recreated_scenario: Scenario):
        assert recreated_scenario.requests_count == 2
        assert recreated_scenario.name == created_scenario.name
        assert recreated_scenario.description == created_scenario.description

      def test_it_creates_requests(self, created_scenario_requests: List[Request], recreated_scenario: Scenario):
        requests = recreated_scenario.requests
        assert len(requests) == 2

        assert_orm_request_equivalent(requests[0], created_scenario_requests[0])
        assert_orm_request_equivalent(requests[1], created_scenario_requests[1])

    class TestWhenMergingScenario():

      class TestWhenAdditionalRequestExists():
        '''
        1. Create scenario
        2. Add 2 requests to it
        3. Snapshot
        4. Add 1 request to it
        5. Apply
        6. Expect scenario to have first 2 requests
        '''

        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def created_scenario_requests(self, runner: CliRunner, created_scenario: Scenario):
          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), NON_DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          return created_scenario.requests

        @pytest.fixture(scope='class', autouse=True)
        def put_event(self, runner: CliRunner, created_scenario: Scenario):
          snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
          assert snapshot_result.exit_code == 0

          log = Log()
          events = log.events
          return events[len(events) - 1]

        # This request was created after the snapshot
        # Later when we apply changes, we expect this request to not exist
        @pytest.fixture(scope='class', autouse=True)
        def new_scenario_request(self, runner: CliRunner, created_scenario: Scenario, put_event: LogEvent):
          assert created_scenario.uuid == put_event.resource_uuid
          assert created_scenario.requests.count() == 2

          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          return created_scenario.requests.last()

        @pytest.fixture(scope='class')
        def apply_result(self, runner: CliRunner, created_scenario: Scenario):
          created_scenario = Scenario.find(created_scenario.id)
          assert created_scenario.requests_count == 3
          apply_result = runner.invoke(snapshot, ['apply'])

          return apply_result

        def test_it_merges_scenario(self, created_scenario: Scenario, apply_result):
          assert apply_result.exit_code == 0

          created_scenario = Scenario.find(created_scenario.id)
          assert created_scenario.requests_count == 2

        def test_it_maintains_requests(self, created_scenario: Scenario, created_scenario_requests: List[Request]):
          created_scenario = Scenario.find(created_scenario.id)
          assert len(created_scenario.requests) == 2

          requests = created_scenario.requests

          assert_orm_request_equivalent(requests[0], created_scenario_requests[0])
          assert_orm_request_equivalent(requests[1], created_scenario_requests[1])

      class TestWhenRequestsDeleted():
        '''
        1. Create scenario
        2. Add 2 requests to it
        3. Snapshot
        4. Delete last request added
        5. Snapshot
        6. Re-create last request
        7. Apply
        8. Expect scenario to have 1 request
        '''

        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def created_scenario_requests(self, runner: CliRunner, created_scenario: Scenario):
          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), NON_DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          return created_scenario.requests

        @pytest.fixture(scope='class', autouse=True)
        def delete_event(self, runner: CliRunner, created_scenario: Scenario):
          snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
          assert snapshot_result.exit_code == 0

          last_request = created_scenario.requests.last()
          last_request.delete()

          snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
          assert snapshot_result.exit_code == 0

          # Re-create the request
          Request.create(**last_request.attributes_to_dict())

          log = Log()
          events = log.events
          return events[len(events) - 1]

        @pytest.fixture(scope='class')
        def apply_result(self, runner: CliRunner, created_scenario: Scenario):
          created_scenario = Scenario.find(created_scenario.id)
          apply_result = runner.invoke(snapshot, ['apply'])

          return apply_result

        def test_it_merges_scenario(self, created_scenario: Scenario, apply_result):
          assert apply_result.exit_code == 0

          created_scenario = Scenario.find(created_scenario.id)
          assert created_scenario.requests_count == 1

        def test_it_maintains_requests(self, created_scenario: Scenario, created_scenario_requests: List[Request]):
          created_scenario = Scenario.find(created_scenario.id)
          requests = created_scenario.requests

          assert_orm_request_equivalent(requests[0], created_scenario_requests[0])

        def test_it_removes_request_snapshot(self, delete_event: LogEvent):
          scenario_snapshot = ScenarioSnapshot(delete_event.resource_uuid)
          assert len(scenario_snapshot.requests) == 1

      class TestWhenMergingRequests():
        '''
        1. Create 2 requests with event id A and B respectively
        2. Snapshot
        3. Apply snapshots so that log VERSION is [A, B]
        4. Remove A from VERSION, expect VERSION is [B]
        5. Apply snapshots
        6. Expect log VERSION is again [A, B]
        '''

        @pytest.fixture(scope='class')
        def recorded_request_one(self, runner: CliRunner):
          record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0
          return Request.last()

        @pytest.fixture(scope='class')
        def recorded_request_two(self, runner: CliRunner):
          record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0
          return Request.last()

        @pytest.fixture(scope='class', autouse=True)
        def snapshot_one_result(self, runner: CliRunner, recorded_request_one: Request):
          snapshot_result = runner.invoke(request, ['snapshot', recorded_request_one.key()])
          assert snapshot_result.exit_code == 0
          return snapshot_result

        @pytest.fixture(scope='class', autouse=True)
        def snapshot_two_result(self, runner: CliRunner, recorded_request_two: Request):
          snapshot_result = runner.invoke(request, ['snapshot', recorded_request_two.key()])
          assert snapshot_result.exit_code == 0
          return snapshot_result

        @pytest.fixture()
        def apply_result(self, runner: CliRunner):
          apply_result = runner.invoke(snapshot, ['apply'])
          assert apply_result.exit_code == 0
          return apply_result

        def test_it_applies_older_snapshot(self, runner: CliRunner, apply_result):
          log = Log()
          expected_version = log.version

          events = log.events
          log.version = events[1].uuid
          assert log.version != expected_version

          apply_result = runner.invoke(snapshot, ['apply'])
          assert apply_result.exit_code == 0

          assert log.version == expected_version

    class TestWhenMalformedScenarioMetadata():

      @pytest.fixture(scope='class')
      def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

      @pytest.fixture(scope='class', autouse=True)
      def put_event(self, runner: CliRunner, created_scenario: Scenario):
        snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        event = events[len(events) - 1]
        metadata_path = event.snapshot().metadata_path
        with open(metadata_path, 'w') as fp:
          fp.write('junk')

        return event

      @pytest.fixture(scope='class', autouse=True)
      def removed_scenario(self, created_scenario: Scenario):
        created_scenario.delete()
        assert Scenario.find(created_scenario.id) == None

      def test_creates_scenario_with_uuid_name(self, runner: CliRunner, created_scenario: Scenario):
        apply_result = runner.invoke(snapshot, ['apply'])
        assert apply_result.exit_code == 0

        recreated_scenario = Scenario.find_by(uuid=created_scenario.uuid)
        assert recreated_scenario.name == created_scenario.uuid

    class TestWhenMissingScenarioMetadata():

      @pytest.fixture(scope='class')
      def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

      @pytest.fixture(scope='class', autouse=True)
      def put_event(self, runner: CliRunner, created_scenario: Scenario):
        snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        event = events[len(events) - 1]
        metadata_path = event.snapshot().metadata_path
        os.remove(metadata_path)

        return event

      @pytest.fixture(scope='class', autouse=True)
      def removed_scenario(self, created_scenario: Scenario):
        created_scenario.delete()
        assert Scenario.find(created_scenario.id) == None

      def test_skips_scenario(self, runner: CliRunner, created_scenario: Scenario):
        apply_result = runner.invoke(snapshot, ['apply'])
        assert apply_result.exit_code == 0

        recreated_scenario = Scenario.find_by(uuid=created_scenario.uuid)
        assert recreated_scenario == None
  
    class TestWhenMalformedRequest():

      @pytest.fixture(scope='class')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(autouse=True, scope='class')
      def put_event(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        event = events[len(events) - 1]

        with open(event.snapshot().path, 'w') as fp:
          fp.write('junk')

        return event

      @pytest.fixture(scope='class', autouse=True)
      def removed_request(self, recorded_request: Request):
        recorded_request.delete()
        assert Request.find(recorded_request.id) == None

      def test_it_creates_request(self, runner: CliRunner, recorded_request: Request):
        apply_result = runner.invoke(snapshot, ['apply'])
        assert apply_result.exit_code == 1

        assert Request.find_by(uuid=recorded_request.uuid) == None

      def test_idempotent(self, runner: CliRunner, recorded_request: Request, put_event: LogEvent):
        apply_result = runner.invoke(snapshot, ['apply', put_event.uuid])
        assert apply_result.exit_code == 1

        assert Request.find_by(uuid=recorded_request.uuid) == None

    class TestWhenMalformedScenarioRequest():

      @pytest.fixture(scope='class')
      def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

      @pytest.fixture(scope='class', autouse=True)
      def created_scenario_requests(self, runner: CliRunner, created_scenario: Scenario):
          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0

          return created_scenario.requests

      @pytest.fixture(scope='class', autouse=True)
      def put_event(self, runner: CliRunner, created_scenario: Scenario):
        snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        event = events[len(events) - 1]
        event.snapshot().iter_request_snapshots(lambda snapshot: write_junk(snapshot.path))

        return event

      @pytest.fixture(scope='class', autouse=True)
      def removed_scenario(self, created_scenario: Scenario):
        created_scenario.delete()
        assert Scenario.find(created_scenario.id) == None

      def test_skips_scenario(self, runner: CliRunner, created_scenario: Scenario):
        apply_result = runner.invoke(snapshot, ['apply'])
        assert apply_result.exit_code == 1

        recreated_scenario = Scenario.find_by(uuid=created_scenario.uuid)
        assert recreated_scenario

        assert len(recreated_scenario.requests) == 0

      def test_idempotent(self, runner: CliRunner, created_scenario: Scenario, put_event: LogEvent):
        apply_result = runner.invoke(snapshot, ['apply', put_event.uuid])
        assert apply_result.exit_code == 1

        recreated_scenario = Scenario.find_by(uuid=created_scenario.uuid)
        assert recreated_scenario

        assert len(recreated_scenario.requests) == 0


import pdb

from stoobly_agent.app.proxy.replay.body_parser_service import decode_response

from stoobly_agent.config.constants import test_strategy

from .context_abc import TestContextABC as TestContext
from .matchers.custom import matches as custom_matches
from .matchers.contract import matches as contract_matches
from .matchers.diff import matches as diff_matches
from .matchers.fuzzy import matches as fuzzy_matches

def test(context: TestContext):
    if context.skipped:
        return None, ''

    valid_test_strategy, log = __validate_test_strategy(context)
    if not valid_test_strategy:
        return False, log
    
    # Test contract
    contract_matches, contract_log = __test_request_contract(context)

    # Test status code
    status_code_matches, status_code_log = __test_status_code(context)
    
    # Test response
    response_matches, log = __test_response(context) 

    # Format test results
    log_lines = []

    if not contract_matches:
        log_lines.append(f"\n=== Contract Test ===")
        log_lines.append(contract_log)

    if not status_code_matches:
        log_lines.append(f"\n=== Status Code Test ===")
        log_lines.append(status_code_log)

    if not response_matches:
        log_lines.append(f"\n=== Response Test ===")
        log_lines.append(log)

    context.log = "\n".join(log_lines).lstrip()
    context.passed = status_code_matches and contract_matches and response_matches 

    return context.passed, context.log

def __validate_test_strategy(context: TestContext):
    valid_strategies = [test_strategy.CONTRACT, test_strategy.CUSTOM, test_strategy.DIFF, test_strategy.FUZZY]
    active_test_strategy = context.strategy

    if active_test_strategy not in valid_strategies:
        test_strategies = ','.join(valid_strategies)
        return False, f"Could not find matching test strategy: valid options [{test_strategies}]"

    return True, ''

# Tests

def __test_request_contract(context: TestContext):
    endpoint = context.endpoint

    if not endpoint:
        return True, ''
 
    request = context.flow.request

    headers = context.request_headers
    normalized_headers = {k.lower(): v for k, v in headers.items()} # Headers are case insensitive
    header_names_facade = endpoint.header_names
 
    matches, log = contract_matches(context, header_names_facade, normalized_headers)
    if not matches:
        return matches, f"{log} in headers"

    query_params = request.query
    query_param_names_facade = endpoint.query_param_names
    matches, log = contract_matches(context, query_param_names_facade, query_params, strict=True)
    if not matches:
        return matches, f"{log} in query params" 

    body_params = decode_response(request.content, headers.get('content-type'))
    body_param_names_facade = endpoint.body_param_names
    matches, log = contract_matches(context, body_param_names_facade, body_params, strict=True)
    if not matches:
        return matches, f"{log} in body params" 

    return True, ''

# TODO
def __test_response_headers(context):
    return True, ''

def __test_response(context: TestContext):
    active_test_strategy = context.strategy
    content = context.decoded_response_content
    expected_content = context.rewritten_expected_response_content
    response_param_names_facade = context.response_param_names

    if active_test_strategy == test_strategy.DIFF: 
        return diff_matches(context, response_param_names_facade, expected_content, content)
    elif active_test_strategy == test_strategy.FUZZY:
        return fuzzy_matches(context, response_param_names_facade, expected_content, content)
    elif active_test_strategy == test_strategy.CUSTOM:
        return custom_matches(context)
    elif active_test_strategy == test_strategy.CONTRACT:
        return contract_matches(context, response_param_names_facade, content)
    else:
        raise 'Setting Error: missing test test strategy'

def __test_status_code(context: TestContext) -> bool:
    response = context.response
    expected_response = context.expected_response

    matches = response.status_code == expected_response.status_code 

    log = ''
    if not matches:
        log = f"Status codes did not match: got {response.status_code} expected {expected_response.status_code}"

    return matches, log


import pdb
import requests

from io import BytesIO
from urllib3 import HTTPResponse
from typing import TypedDict, Union

from stoobly_agent.app.models.helpers.create_request_params_service import build_params_from_python
from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.proxy.replay.body_parser_service import encode_response
from stoobly_agent.app.settings import Settings

class RequestProperties(TypedDict):
  url: str
  method: str
  request_body: bytes
  request_headers: dict
  response_body: bytes
  response_headers: bytes
  status_code: int

class RequestBuilder():

  def __init__(self, **properties: RequestProperties):
    self.__properties = properties
    self.__settings = None

  def with_settings(self, settings: Settings):
    self.__settings = settings
    return self

  @property
  def method(self):
    return self.__properties.get('method').upper()

  @property
  def request_body(self):
    return self.__properties.get('request_body') or b''

  @property
  def request_headers(self):
    return self.__properties.get('request_headers') or {}

  @property
  def response_body(self):
    return self.__properties.get('response_body')

  @property
  def response_headers(self):
    return self.__properties.get('response_headers') or {}

  @property
  def settings(self):
    return self.__settings or Settings.instance()

  @property
  def status_code(self):
    return self.__properties.get('status_code') or 0

  @property
  def url(self):
    return self.__properties.get('url')

  def build(self):
    req = requests.Request(
      method=self.method,
      url=self.url,
      headers=self.request_headers,
      data=self.serialized_request_body(self.request_headers_value('content-type')),
    )
    
    res = requests.Response()
    res.headers = self.response_headers
    res.raw = HTTPResponse(
        body=BytesIO(
          self.serialized_response_body(self.response_headers_value('content-type')),
        ),
        decode_content=False,
        headers=self.response_headers,
        preload_content=False
    ) 
    res.status_code = self.status_code

    params = build_params_from_python(req, res)

    model = RequestModel(self.settings)
    return model.create(**params)

  def request_headers_value(self, header_name: str):
    headers = self.request_headers
    return self.__header_value(header_name, headers)

  def response_headers_value(self, header_name: str):
    headers = self.response_headers
    return self.__header_value(header_name, headers)

  def serialized_request_body(self, content_type: str):
    body = self.request_body
    return self.__serialize_body(body, content_type)

  def serialized_response_body(self, content_type: str):
    body = self.response_body
    return self.__serialize_body(body, content_type)

  def __header_value(self, header_name: str, headers: dict):
    for k in headers:
      if header_name.lower() == k.lower():
        return headers[k]

  def __serialize_body(self, body: Union[bytes, str], content_type: str):
    if not isinstance(body, bytes) or not isinstance(body, str): 
      body = encode_response(body, content_type)
    return body if isinstance(body, bytes) else body.encode()




import pdb
import pytest

from click.testing import CliRunner
from pathlib import Path

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL

from stoobly_agent.config.constants.custom_headers import SESSION_ID
from stoobly_agent.cli import mock, record, scenario
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

class TestMocking():
  @pytest.fixture(scope='class')
  def lifecycle_hooks_path(self):
    return str(Path(__file__).parent / 'mock_scenario_lifecycle_hooks.py')

  class TestScenario():
    @pytest.fixture(scope='class')
    def scenario_key(self, runner: CliRunner):
      res = runner.invoke(scenario, ['create', '--select', 'key', '--without-headers', 'test-scenario'])
      assert res.exit_code == 0
      return ScenarioKey(res.stdout.strip()).raw

    @pytest.fixture(autouse=True, scope='class')
    def recorded_request1(self, runner: CliRunner, scenario_key: str):
      record_result = runner.invoke(record, ['--scenario-key', scenario_key, DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      return Request.last()

    @pytest.fixture(autouse=True, scope='class')
    def recorded_request2(self, runner: CliRunner, scenario_key: str):
      record_result = runner.invoke(record, ['--scenario-key', scenario_key, DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      return Request.last()

    def test_it_does_not_mocks(self, runner: CliRunner, scenario_key: str):
      mock_result = runner.invoke(mock, ['--scenario-key', scenario_key, NON_DETERMINISTIC_GET_REQUEST_URL])
      assert mock_result.exit_code == 1

    def test_it_mocks(self, runner: CliRunner, scenario_key: str):
      mock_result = runner.invoke(mock, ['--scenario-key', scenario_key, DETERMINISTIC_GET_REQUEST_URL])
      assert mock_result.exit_code == 0

    def test_it_mocks_in_order(
      self, runner: CliRunner, lifecycle_hooks_path: str, recorded_request1: Request, recorded_request2: Request, scenario_key: str
    ):
      session_id = 'test'
      args = [
        '--lifecycle-hooks-path', lifecycle_hooks_path, '--scenario-key', scenario_key, '--output', '/dev/null',
        '-H', f"{SESSION_ID}: {session_id}", DETERMINISTIC_GET_REQUEST_URL
      ]
      mock_result = runner.invoke(mock, args)
      assert int(mock_result.stdout) == recorded_request1.id

      mock_result = runner.invoke(mock, args)
      assert int(mock_result.stdout) == recorded_request2.id


import pdb

from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.config.constants import test_filter
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse

class EndpointFacade():

  def __init__(self, endpoints_resource: EndpointsResource, endpoint_id: int):
    self.__endpoints_resource = endpoints_resource
    self.__endpoint_id = endpoint_id
    
    self.trace = None
    self.filter = test_filter.ALL

    # Cache
    self.__cached_body_param_names = None
    self.__cached_endpoint_show_response = None
    self.__cached_header_names = None
    self.__cached_query_param_names = None
    self.__cached_response_param_names = None

  def get(self) -> EndpointShowResponse:
    if self.__cached_endpoint_show_response:
      return self.__cached_endpoint_show_response

    endpoint_id = self.__endpoint_id
    resource = self.__endpoints_resource

    if not resource or not endpoint_id:
      return 

    try:
      res = resource.show(
        endpoint_id,
        aliases=True, 
        header_names=True,
        query_param_names=True,
        body_param_names=True,
        response_param_names=True
      )
    except Exception as e:
      return 

    if not res.ok:
      return 

    self.__cached_endpoint_show_response: EndpointShowResponse = res.json()

    return self.__cached_endpoint_show_response

  def with_show_response(self, res: EndpointShowResponse):
    self.__cached_endpoint_show_response = res
    return self

  @property
  def aliases(self):
    endpoint_show_response: EndpointShowResponse = self.get()
    return endpoint_show_response.get('aliases', [])

  @property
  def header_names(self) -> RequestComponentNamesFacade:
    if self.__cached_header_names:
      return self.__cached_header_names

    self.__cached_header_names = self.__request_component_names('header_names')

    return self.__cached_header_names

  @property
  def query_param_names(self) -> RequestComponentNamesFacade:
    if self.__cached_query_param_names:
      return self.__cached_query_param_names

    self.__cached_query_param_names = self.__request_component_names('query_param_names')

    return self.__cached_query_param_names

  @property
  def body_param_names(self) -> RequestComponentNamesFacade:
    if self.__cached_body_param_names:
      return self.__cached_body_param_names

    self.__cached_body_param_names = self.__request_component_names('body_param_names')

    return self.__cached_body_param_names

  @property
  def response_param_names(self) -> RequestComponentNamesFacade:
    if self.__cached_response_param_names:
      return self.__cached_response_param_names

    self.__cached_response_param_names = self.__request_component_names('response_param_names')

    return self.__cached_response_param_names

  def __request_component_names(self, component_name):
    endpoint_show_response: EndpointShowResponse = self.get()
    aliases = self.aliases
    component_names = endpoint_show_response.get(component_name) or []

    return RequestComponentNamesFacade(
      component_names      
    ).with_aliases(aliases).with_filter(self.filter).with_trace(self.trace)



import click
import json
import pdb
import sys

from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias
from stoobly_agent.lib.orm.trace_request import TraceRequest

from .helpers.print_service import FORMATS, print_traces, select_print_options
from .helpers.validations import *

@click.group(
    epilog="Run 'stoobly-agent trace COMMAND --help' for more information on a command.",
    help="Manage traces"
)
@click.pass_context
def trace(ctx):
    pass

@trace.command(
    help="Create a trace"
)
def create(**kwargs):
    trace = Trace.create()
    print(trace.id)

@trace.command(
  help="Clone a trace alias"
)
@click.argument('trace_id')
def clone(**kwargs):
  trace = Trace.find_by(id=kwargs['trace_id'])
  if not trace:
    print(f"Could not find trace {trace.id}", file=sys.stderr)
    sys.exit(1)

  cloned_trace = trace.clone()
  if not cloned_trace:
    print(f"Could not clone trace {trace.id}", file=sys.stderr)
    sys.exit(1)

  print(cloned_trace.id)

@trace.command(
    help="Show all traces"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--size', default=10)
@click.option('--sort-by', default='created_at', help='created_at|name')
@click.option('--sort-order', default='desc', help='asc | desc')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
def list(**kwargs):
  print_options = select_print_options(kwargs)

  traces = Trace.limit(kwargs['size']).order_by(kwargs['sort_by'], kwargs['sort_order'])

  traces_json = json.loads(traces.get().to_json())
  for trace_json in traces_json:
    id = trace_json['id']

    trace_requests = TraceRequest.where_for(trace_id=id)
    trace_json['trace_requests_count'] = trace_requests.count()

    trace_aliases = TraceAlias.where_for(trace_id=id)
    trace_json['trace_aliases_count'] = trace_aliases.count()

  print_traces(traces_json, **{ **print_options, 'filter': ['updated_at']})

@click.group(
  epilog="Run 'stoobly-agent trace alias COMMAND --help' for more information on a command.",
  help="Manage trace aliases",
)
@click.pass_context
def alias(ctx):
  pass

@alias.command(
    help="Show all aliases for a trace"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--name', help='Filter by name.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--size', default=10)
@click.option('--sort-by', default='created_at', help='created_at|name')
@click.option('--sort-order', default='desc', help='asc | desc')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.argument('trace_id')
def list(**kwargs):
  print_options = select_print_options(kwargs)

  columns = {
    'trace_id': kwargs['trace_id']
  }

  if kwargs['name']:
    columns['name'] = kwargs['name']

  trace_aliases = TraceAlias.where_for(**columns)
  trace_aliases = trace_aliases.limit(kwargs['size']).order_by(kwargs['sort_by'], kwargs['sort_order'])

  print_traces(json.loads(trace_aliases.get().to_json()), **{ **print_options, 'filter': ['trace_id', 'updated_at']})

@alias.command(
  help="Create a trace alias"
)
@click.option('--name', required=True)
@click.option('--value', required=True)
@click.argument('trace_id')
def create(**kwargs):
  try:
    value = eval(kwargs['value'])
  except Exception as e:
    value = kwargs['value']

  trace_alias = TraceAlias.create(trace_id=kwargs['trace_id'], name=kwargs['name'], value=value)
  print(trace_alias.id)

@alias.command(
  help="Show a trace alias"
)
@click.option('--name', required=True)
@click.argument('trace_id')
def show(**kwargs):
  trace_alias = TraceAlias.find_by(trace_id=kwargs['trace_id'], name=kwargs['name'])

  if not trace_alias:
    print('Not found', file=sys.stderr)
    sys.exit(1)
  else:
    sys.stdout.write(trace_alias.value)

@alias.command(
  help="Update a trace alias"
)
@click.option('--name', required=True)
@click.option('--value', required=True)
@click.argument('trace_id')
def update(**kwargs):
  try:
    value = eval(kwargs['value'])
  except Exception as e:
    value = kwargs['value']

  trace_alias = TraceAlias.find_by(trace_id=kwargs['trace_id'], name=kwargs['name'])
  if not trace_alias:
    trace_alias = TraceAlias.create(trace_id=kwargs['trace_id'], name=kwargs['name'], value=value)
  else:
    trace_alias.update(value=value)
  
  print(trace_alias.id)

trace.add_command(alias)


import pdb

from copy import deepcopy
from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.proxy.replay.body_parser_service import encode_response
from stoobly_agent.app.proxy.replay.context import ReplayContext
from stoobly_agent.app.proxy.utils.request_handler import build_response
from stoobly_agent.app.proxy.utils.response_handler import disable_transfer_encoding
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import custom_headers, lifecycle_hooks, request_origin
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.interfaces.tests import TestShowResponse
from stoobly_agent.lib.logger import Logger

from .handle_mock_service import handle_request_mock_generic_without_rewrite
from .handle_replay_service import handle_request_replay_without_rewrite
from .mock.context import MockContext
from .test.helpers.test_results_builder import TestResultsBuilder
from .test.helpers.upload_test_service import inject_upload_test
from .test.context_abc import TestContextABC as TestContext
from .test.test_service import test
from .utils.rewrite import rewrite_request, rewrite_response, rewrite_request_response

LOG_ID = 'HandleTest'

###
#
# 1. Rewrites test request
# 2. BEFORE_REPLAY gets triggered
#
def handle_request_test(context: ReplayContext) -> None:
    __rewrite_request(context)
    handle_request_replay_without_rewrite(context)

###
#
# 1.  Rewrites test response (response from live service)
# 2.  AFTER_REPLAY gets triggered
# 3.  Uses rewritten test request to obtain mock response
# 4.  BEFORE_MOCK gets triggered
# 5.  AFTER_MOCK gets triggered
# 6.  BEFORE_TEST gets triggered
# 7.  Tests against rewritten test response and mock response (expected response)
# 8.  Rewrites a copy of request and response
# 9.  BEFORE_RECORD gets triggered (if not from CLI)
# 10. AFTER_RECORD gets triggered (if not from CLI)
# 11. AFTER_TEST gets triggered 
#
def handle_response_test(context: ReplayContext) -> None:
    from .test.context import TestContext

    flow: MitmproxyHTTPFlow = context.flow
    disable_transfer_encoding(flow.response)

    __rewrite_response(context)
    __test_hook(lifecycle_hooks.AFTER_REPLAY, context)

    intercept_settings = context.intercept_settings

    # At this point, the request may already been rewritten during replay, do not rewrite again
    handle_request_mock_generic_without_rewrite(
        MockContext(flow, intercept_settings),
        error=lambda mock_context: __handle_mock_error(TestContext(context, mock_context)),
        failure=lambda mock_context: __handle_mock_failure(TestContext(context, mock_context)),
        #infer=intercept_settings.test_strategy == test_strategy.FUZZY, # For fuzzy testing we can use an inferred response
        success=lambda mock_context: __handle_mock_success(TestContext(context, mock_context))
    )

def __decorate_test_id(flow: MitmproxyHTTPFlow, test_response: TestShowResponse):
    if test_response.get('id'):
        flow.response.headers[custom_headers.TEST_ID] = str(test_response['id'])

def __handle_mock_error(test_context: TestContext):
    Logger.instance().warn(f"{LOG_ID}:TestStatus: Mock not enabled")

    intercept_settings = test_context.intercept_settings

    if intercept_settings.request_origin == request_origin.CLI:
        return build_response(False, 'No test found')

def __handle_mock_failure(test_context: TestContext) -> None:
    Logger.instance().warn(f"{LOG_ID}:TestStatus: No test found")

    intercept_settings = test_context.intercept_settings

    if intercept_settings.request_origin == request_origin.CLI:
        return build_response(False, 'No test found')

def __handle_mock_success(test_context: TestContext) -> None:
    flow: MitmproxyHTTPFlow = test_context.flow

    __test_hook(lifecycle_hooks.AFTER_MOCK, test_context.mock_context)

    settings: Settings = Settings.instance()
    test_context.with_endpoints_resource(EndpointsResource(settings.remote.api_url, settings.remote.api_key))

    __test_hook(lifecycle_hooks.BEFORE_TEST, test_context)

    intercept_settings = test_context.intercept_settings
    if intercept_settings.test_skip:
        passed, log = (False, '')
        skipped = True
    else: 
        # Run test
        passed, log = test(test_context)
        skipped = False

    request_id = test_context.mock_request_id
    if request_id:
        expected = test_context.cached_rewritten_expected_response_content
        upload_test_data = {
            'log': log,
            'passed': passed,
            'request_id': request_id,
            'skipped': skipped,
            'status': flow.response.status_code,
            'strategy': test_context.strategy
        }

        is_cli = intercept_settings.request_origin == request_origin.CLI
        if not is_cli or test_context.save:
            # Re-serialize expected response since it was rewritten
            upload_test_data['expected_response'] = encode_response(expected, test_context.expected_response.content_type)

            res = __record_handler(test_context, upload_test_data) 

            if is_cli:
                # If the origin was from a CLI, send test ID in response header
                if res.ok:
                    __decorate_test_id(flow, res.json())
                else:
                    # If we failed to upload test results, provide a local response
                    Logger.instance().warn(f"{LOG_ID}:TestStatus: Failed to upload results")
                    test_context.save = False

        if is_cli and not test_context.save:
            # No serialization is needed since TestResultsBuilder will serialize
            upload_test_data['expected_response'] = expected
            received = test_context.decoded_response_content

            builder = TestResultsBuilder(
                **{ 
                    'expected_latency': test_context.expected_latency,
                    'expected_status_code': test_context.expected_status_code,
                    'received_response': received
                },
                **upload_test_data
            )

            __override_response(flow, builder.serialize())

    __test_hook(lifecycle_hooks.AFTER_TEST, test_context)
    
    return flow.response

def __override_response(flow: MitmproxyHTTPFlow, content: bytes):
    headers = { 'Content-Type': 'text/plain' }
    headers[custom_headers.CONTENT_TYPE] = custom_headers.CONTENT_TYPE_TEST_RESULTS
    flow.response.headers = headers
    flow.response.set_content(content)
    flow.response.status_code = 200

def __record_handler(context: TestContext, upload_test_data):
    flow = context.flow
    flow_copy = deepcopy(flow) # Deep copy flow to prevent response modifications from persisting
    intercept_settings = context.intercept_settings

    context.flow = flow_copy 

    # Since we are "uploading" the request, use record_write_rules
    rewrite_request_response(flow_copy, intercept_settings.record_rewrite_rules)
    __test_hook(lifecycle_hooks.BEFORE_RECORD, context)

    # Commit test to API
    upload_test = inject_upload_test(None, intercept_settings)
    res = upload_test(
        flow_copy, **upload_test_data
    )

    __test_hook(lifecycle_hooks.AFTER_RECORD, context)
    context.flow = flow

    return res

def __rewrite_request(replay_context: ReplayContext):
    """
    Before replaying a request, see if the request needs to be rewritten
    """
    intercept_settings: InterceptSettings = replay_context.intercept_settings
    rewrite_rules = intercept_settings.test_rewrite_rules

    if len(rewrite_rules) > 0:
        rewrite_request(replay_context.flow, rewrite_rules)

def __rewrite_response(replay_context: ReplayContext):
    """
    After replaying a request, see if the request needs to be rewritten
    """
    intercept_settings: InterceptSettings = replay_context.intercept_settings
    rewrite_rules = intercept_settings.test_rewrite_rules

    if len(rewrite_rules) > 0:
        rewrite_response(replay_context.flow, rewrite_rules)

def __test_hook(hook: str, context: TestContext):
    intercept_settings = context.intercept_settings
    lifecycle_hooks_module = intercept_settings.lifecycle_hooks

    if hook in lifecycle_hooks_module:
        lifecycle_hooks_module[hook](context)


import base64
import json
import pdb
import requests

from typing import List, Tuple

from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.lib.api.interfaces import Header, HeaderShowResponse
from stoobly_agent.lib.orm.request import Request

from .local_db_adapter import LocalDBAdapter
from .response_adapter import LocalDBResponseAdapter

class LocalDBResponseHeaderAdapter(LocalDBAdapter):
  __request_orm = None

  def __init__(self, request_orm: Request.__class__ = Request):
    self.__request_orm = request_orm

  def create(self, request_id, **header: Header):
    return self.update(request_id, **header)

  def update(self, request_id, **header: Header):
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    response = request.response

    if not response:
      return self.__response_not_found()

    name = header['name']
    value = header['value']

    python_response: requests.Response = RawHttpResponseAdapter(response.raw).to_response()
    headers = python_response.headers
    headers[name] = value

    LocalDBResponseAdapter(self.__request_orm).update(response.id, headers=headers)

    return self.success({
      'name': name,
      'value': value,
    })

  def index(self, response_id, **query_params) -> Tuple[List[HeaderShowResponse], int]:
    request = self.__request_orm.find(response_id)

    if not request:
      return self.__request_not_found()

    response = request.response
    if not response:
      return self.__response_not_found()

    response: requests.Response = RawHttpResponseAdapter(response.raw).to_response()

    headers = []
    for key, val in response.headers.items():
      headers.append({
        'name': key,
        'value': val,
      })

    return self.success(headers)

  def destroy(self, request_id, id):
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    response = request.response
    if not response:
      return self.__response_not_found()

    decoded_id = self.__decode_id(id)
    if not decoded_id:
      return self.bad_request('Invalid id')

    response: requests.Request = RawHttpResponseAdapter(response.raw).to_response()
    headers = response.headers

    header_name = decoded_id['name'] 
    if not header_name in headers:
      return self.not_found('Response header not found')

    del headers[header_name]

    LocalDBResponseAdapter(self.__request_orm).update(request_id, headers=headers)

    return self.success({
      'name': decoded_id['name'],
    })

  def __decode_id(self, id: str):
    id = base64.b64decode(id)

    try:
      id = json.loads(id)
    except Exception as e:
      return None

    if not isinstance(id, dict):
      return None

    if not id.get('name'):
      return None

    return id

  def __request_not_found(self):
    return self.not_found('Request not found')

  def __response_not_found(self):
    return self.not_found('Response not found')


import pdb

from typing import List

from .match_rule import MatchRule

from .types.proxy_settings import MatchSettings as IMatchSettings

class MatchSettings:
  __match_settings = None

  def __init__(self, match_settings: IMatchSettings):
    self.__match_settings = match_settings or {}

    self.__match_rules_map = {
      k: list(map(lambda rule: MatchRule(rule), v)) for k, v in self.__match_settings.items()
    }

  def match_rules(self, project_id: str) -> List[MatchRule]:
    return self.__match_rules_map.get(project_id) or []

  def set_match_rules(self, project_id: str, v: List[MatchRule]):
    self.__match_rules_map[project_id] = v

  def to_dict(self):
    return {k: list(map(lambda rule: rule.to_dict(), v)) for k, v in self.__match_rules_map.items()}


from typing import Callable, TypedDict

from .openapi_endpoint_adapter import OpenApiEndpointAdapter
from stoobly_agent.app.models.types import OPENAPI_FORMAT
from stoobly_agent.lib.api.body_param_names_resource import BodyParamNamesResource
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.header_names_resource import HeaderNamesResource
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse
from stoobly_agent.lib.api.query_param_names_resource import QueryParamNamesResource
from stoobly_agent.lib.api.response_param_names_resource import ResponseParamNamesResource
from stoobly_agent.lib.api.response_header_names_resource import ResponseHeaderNamesResource

class EndpointsImportContext:
  def __init__(self):
    self.endpoints = []
    self.endpoint_handlers = []
    self.resources = {}
    self.project_id = None

  def with_endpoint_handler(self, handler: Callable[[EndpointShowResponse], None]):
    self.endpoint_handlers.append(handler)
    return self
  
  def with_project(self, project_id: int):
    self.project_id = project_id
    return self
  
  def with_endpoint_resource(self, resource: EndpointsResource):
    self.resources['endpoint'] = resource
    return self
  
  def with_header_name_resource(self, resource: HeaderNamesResource):
    self.resources['header_name'] = resource
    return self

  def with_body_param_name_resource(self, resource: BodyParamNamesResource):
    self.resources['body_param_name'] = resource
    return self

  def with_query_param_name_resource(self, resource: QueryParamNamesResource):
    self.resources['query_param_name'] = resource
    return self

  def with_response_param_name_resource(self, resource: ResponseParamNamesResource):
    self.resources['response_param_name'] = resource
    return self
  
  def with_response_header_name_resource(self, resource: ResponseHeaderNamesResource):
    self.resources['response_header_name'] = resource
    return self

  def with_source(self, path: str, format: str):
    if format == OPENAPI_FORMAT:
      endpoint_adapter = OpenApiEndpointAdapter()
      self.endpoints += endpoint_adapter.adapt_from_file(path)

    return self


import os
import pdb
import requests

from stoobly_agent.app.models.adapters.python.request import PythonRequestAdapterFactory
from stoobly_agent.app.proxy.mitmproxy.request_facade import MitmproxyRequestFacade
from stoobly_agent.app.proxy.record.proxy_request import ProxyRequest
from stoobly_agent.app.proxy.record.request_string import RequestString
from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter

class TestRequestString():
    
    def test_www_google_com(self):
        request = requests.Request(
            data=b'',
            headers={
                'Content-Length': '0'
            },
            method='GET',
            url='https://www.google.com',
        )

        request_string = self.__to_request_string(request)
        raw_request_string = request_string.get()
        _request = RawHttpRequestAdapter(raw_request_string).to_request()

        self.__test_equivalence(_request, request)

    def test_stoobly_com(self):
        data = b'{"name": "test", "description": "test"}'
        request = requests.Request(
            data=data,
            headers={
                'Content-Length': str(len(data)),
                'Content-Type': 'application/json',
            },
            method='POST',
            url='https://stoobly.com/scenarios',
        )

        request_string = self.__to_request_string(request)
        raw_request_string = request_string.get()
        _request = RawHttpRequestAdapter(raw_request_string).to_request()

        self.__test_equivalence(_request, request)

    def __test_equivalence(self, request: requests.Request, _request: requests.Request):
        assert _request.method == request.method

        # MitmproxyRequest adds a trailing /
        assert _request.url.rstrip('/') == request.url.rstrip('/')

        assert _request.data == request.data

        for key, val in _request.headers.items():
            assert val == request.headers.get(key)

    def __to_request_string(self, request: requests.Request):
        mitmproxy_request = PythonRequestAdapterFactory(request).mitmproxy_request()
        mitmproxy_request_facade = MitmproxyRequestFacade(mitmproxy_request)
        proxy_request = ProxyRequest(mitmproxy_request_facade)
        return RequestString(proxy_request)


import os
import pdb
import re

from mergedeep import merge
from mitmproxy.coretypes.multidict import MultiDict
from urllib.parse import urlparse, parse_qs

from stoobly_agent.app.proxy.replay.body_parser_service import decode_response
from stoobly_agent.config.constants import custom_headers, headers

from .routes import ROUTES
from .simple_http_request_handler import SimpleHTTPRequestHandler

class ApplicationHTTPRequestHandler(SimpleHTTPRequestHandler):
    ROOT_DIR = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..'))

    @property
    def public_dir(self):
        return os.path.join(self.ROOT_DIR, 'public')

    ### Method handlers

    def do_OPTIONS(self):
        self.render(
            plain = 'OK',
            status = 200
        )

    def do_POST(self):
        self.preprocess()

        self.body = self.parse_body()
        if isinstance(self.body, dict) or isinstance(self.body, MultiDict):
            merge(self.params, self.body) 
        
        if not self.route('POST'):
            self.render(
                plain = 'NOT FOUND',
                status = 404
            )

    def do_GET(self):
        self.preprocess()

        merge(self.params, self.parse_query_params())

        if not self.route('GET'):
            public_dir = self.public_dir
            path = os.path.join(public_dir, self.path[1:] if self.path != '/' else 'index.html')

            if not os.path.normpath(path).startswith(public_dir):
                return self.not_found()

            if not os.path.exists(path):
                path = os.path.join(public_dir, 'index.html')

            self.render(
                file = path,
                status = 200
            )

    def do_PUT(self):
        self.preprocess()

        self.body = self.parse_body()
        if isinstance(self.body, dict) or isinstance(self.body, MultiDict):
            merge(self.params, self.body) 

        if not self.route('PUT'):
            self.render(
                plain = 'NOT FOUND',
                status = 404
            )

    def do_DELETE(self):
        self.preprocess()

        if not self.route('DELETE'):
            self.render(
                plain = 'NOT FOUND',
                status = 404
            )

    ### Helpers

    # Override
    def enable_cors(self, _headers = {}):
        rendered_headers = []

        header = 'Access-Control-Allow-Origin'
        self.send_header(header, _headers.get(header) or '*')
        rendered_headers.append(header)

        if self.command == 'OPTIONS':
            header = 'Access-Control-Allow-Methods'
            self.send_header(header, _headers.get(header) or 'GET, OPTIONS, POST, PATCH, PUT, DELETE')
            rendered_headers.append(header)

            header = 'Access-Control-Allow-Headers'
            allowed_headers = _headers.get(header) or ', '.join([
                'Content-Type',
                headers.ACCESS_TOKEN.title(),
                headers.CLIENT.title(),
                custom_headers.DO_PROXY.title(),
                headers.EXPIRY.title(),
                headers.PROXY_HEADERS.title(),
                headers.REQUEST_PATH.title(),
                custom_headers.SERVICE_URL.title(),
                headers.TOKEN_TYPE.title(),
                headers.UID.title(),
            ])
            self.send_header(header, allowed_headers)
            rendered_headers.append(header)

            header = 'Access-Control-Max-Age'
            self.send_header(header, _headers.get(header) or '7200')
            rendered_headers.append(header)

        return rendered_headers

    def preprocess(self):
        self.uri = urlparse(self.path)

        self.fullpath = self.path
        self.path = self.uri.path
        self.params = {}
        self.body = ''

    def route(self, method):
        for endpoint_handler in ROUTES[method]:
            path = endpoint_handler[0]

            matches = self.path == path if isinstance(path, str) else bool(re.match(path, self.path))

            if matches:
                handler = endpoint_handler[1]
                handler(self)
                return True

        return False

    def parse_path_params(self, path_params_map):
        path = self.uri.path
        path_segments = path.split('/')

        if len(path_segments) == 0:
            return

        if path_segments[0] == '':
            path_segments.pop(0)

        for path_param in path_params_map:
            index = path_params_map[path_param]

            try:
                self.params[path_param] = path_segments[index]
            except:
                pass

    def parse_query_params(self):
        query_params = parse_qs(self.uri.query)

        for key, value in query_params.items():
            if len(value) == 1:
                query_params[key] = value[0]

        return query_params

    def parse_body(self):
        if not self.headers.get('Content-Length'):
            body = self.rfile.read()
        else:
            content_length = int(self.headers['Content-Length'])
            body = self.rfile.read(content_length)

        if not self.headers['Content-Type']:
            return body

        return decode_response(body, self.headers['Content-Type'])


import pdb

from requests import Response

from stoobly_agent.config.constants import custom_headers
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter

from ..response import Response as ORMResponse

class ORMToRequestsResponseTransformer():

  def __init__(self, response: ORMResponse):
    self.__body = None
    self.__dirty = False
    self.__headers = None
    self.__response = response
    self.__status = None
    self.__with_response_id = False

  @property
  def dirty(self):
    return self.__dirty

  def transform(self) -> Response:
    adapter = RawHttpResponseAdapter(self.__response.raw)

    if self.__body:
      adapter.body = self.__body 

    if self.__headers:
      adapter.headers = { **adapter.headers, **self.__headers }

    if self.__status:
      adapter.status = self.__status

    python_response = adapter.to_response()

    if self.__with_response_id:
      python_response.headers[custom_headers.RESPONSE_ID] = self.__response.id

    return python_response

  def with_response_id(self):
    self.__with_response_id = True
    return self

  def with_body(self, body: bytes):
    self.__body = body
    self.__dirty = True

    self.with_headers({ 'Content-Length': str(len(body)) })

    return self

  def with_headers(self, headers: dict):
    _headers = self.__headers or {}
    self.__headers = { **_headers, **headers } 
    self.__dirty = True
    return self

  def with_status(self, status: int):
    self.__status = status
    self.__dirty = True
    return self


from stoobly_orator.migrations import Migration

class CreateScenarios(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create('scenarios') as table:
            table.increments('id')
            table.string('name')
            table.integer('requests_count').default(0)
            table.string('description').default("")
            table.boolean('starred').default(False)
            table.integer('priority').default(0)
            table.boolean('is_deleted').default(False)
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop('scenarios')



import json
import pdb
import pytest

from click.testing import CliRunner
from mitmproxy.http import Response as MitmproxyResponse

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.models.adapters.orm.response.mitmproxy_adapter import MitmproxyResponseAdapter
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.orm.response import Response

from stoobly_agent.app.models.factories.resource.local_db.helpers.request_builder import RequestBuilder

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestOrmPythonResponseAdapter():
  @pytest.fixture(scope='class')
  def settings(self):
    return reset()

  @pytest.fixture(scope='class')
  def response_body(self):
    return {
      'first_name': 'John',
      'last_name': 'Smith',
    }

  @pytest.fixture(scope='class')
  def response_headers(self):
    return {
      'Content-Type': 'application/json'
    }

  @pytest.fixture(scope='class')
  def created_response(
    self, settings: Settings, 
    response_headers: dict, response_body: dict
  ):
    status = RequestBuilder(
      method='POST',
      request_body='',
      response_headers=response_headers,
      response_body=response_body,
      status_code=200,
      url='https://petstore.swagger.io/v2/pets',
    ).with_settings(settings).build()[1]
    assert status == 200

    return Response.last()

  @pytest.fixture()
  def mitmproxy_response(self, created_response: Response):
    return MitmproxyResponseAdapter(created_response).adapt()

  def test_it_sets_headers(self, mitmproxy_response: MitmproxyResponse, response_headers: dict):
    data = mitmproxy_response.content
    headers = dict(mitmproxy_response.headers)
    assert headers == {
      **response_headers,
      'Content-Length': str(len(data))
    }

  def test_it_returns_content_as_bytes(self, mitmproxy_response: MitmproxyResponse):
    assert isinstance(mitmproxy_response.content, bytes)

  def test_it_returns_raw_content_as_bytes(self, mitmproxy_response: MitmproxyResponse):
    assert isinstance(mitmproxy_response.raw_content, bytes)

  def test_it_sets_content(self, mitmproxy_response: MitmproxyResponse, response_body: dict):
    content = mitmproxy_response.content
    assert json.dumps(response_body) == content.decode()


import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.cli import record, request
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

@pytest.fixture(scope='class', autouse=True)
def settings():
  return reset()

class TestRequestSnapshot():
  @pytest.fixture(scope='class')
  def recorded_request(self, runner: CliRunner):
    record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
    assert record_result.exit_code == 0
    return Request.last()

  @pytest.fixture(scope='class', autouse=True)
  def snapshot_result(self, runner: CliRunner, recorded_request: Request):
    snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
    assert snapshot_result.exit_code == 0
    return snapshot_result

  def test_it_resets(self, runner: CliRunner, recorded_request: Request):
    original_path = recorded_request.path
    recorded_request.update(path='/test')

    snapshot_result = runner.invoke(request, ['reset', recorded_request.key()])
    assert snapshot_result.exit_code == 0

    _request = Request.find(recorded_request.id)
    assert _request.path == original_path


from typing import TypedDict

class Header(TypedDict):
  name: bytes
  value: bytes

class Response(TypedDict):
  body: bytes
  headers: dict
  status: bytes
  status_code: int
  url: bytes

class ResponseParseHandler():
  __response_shell: Response = None

  def __init__(self, response_shell: dict = {}):
    self.__response_shell = response_shell
    self.__response_shell['headers'] = {}

  def on_body(self, body: bytes):
    self.__response_shell['body'] = body

  def on_header(self, name: bytes, value: bytes):
    self.__response_shell['headers'][name] = value

  def on_status(self, status: bytes):
    self.__response_shell['status'] = status

  def on_url(self, url: bytes):
    self.__response_shell['url'] = url